import pdb

from stoobly_agent.app.cli.helpers.replay_facade import ReplayCliOptions, ReplayFacade, TestCliOptions
from stoobly_agent.app.cli.helpers.iterate_group_by import iterate_group_by
from stoobly_agent.app.models.request_model import RequestModel
from stoobly_agent.app.models.schemas.request import Request
from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.proxy.replay.replay_request_service import ReplayRequestOptions, replay, replay_with_trace
from stoobly_agent.app.proxy.replay.trace_context import TraceContext
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import mode
from stoobly_agent.lib.api.keys import ProjectKey, RequestKey
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey

class RequestFacade(ReplayFacade):

  def __init__(self, settings: Settings):
    self.__model = RequestModel(settings)

    super().__init__(settings)

  def delete(self, request_key: str):
    key = RequestKey(request_key)

    return self.__model.destroy(key.id)

  def show(self, request_key: str, cli_options: dict):
    key = RequestKey(request_key)

    return self.__model.show(key.id, **{ 
      'project_id': key.project_id, 
      **cli_options 
    })

  def snapshot(self, request_key: str, cli_options: dict):
    key = RequestKey(request_key)

    return self.__model.snapshot(key.id, **cli_options)

  def index(self, project_key, cli_options: dict):
    query_params = { **cli_options }

    if project_key:
      key = ProjectKey(project_key)
      query_params['project_id'] = key.id

    if cli_options.get('scenario_key'):
      key = ScenarioKey(cli_options['scenario_key'])
      query_params['scenario_id'] = key.id
    
    return self.__model.index(**query_params)

  def replay(self, request_key: str, cli_options: ReplayCliOptions):
    replay_context = self.__build_replay_context(request_key)
    replay_options = {
      'mode': mode.REPLAY,
      **self.__common_replay_options(request_key),
      **self.common_replay_cli_options(cli_options)
    }

    if cli_options.get('record'):
      replay_options['response_mode'] = mode.RECORD

    trace_context = replay_options.get('trace_context')

    return self.__replay(replay_context, trace_context, replay_options)

  def mock(self, request_key: str, cli_options: ReplayCliOptions):
    replay_context = self.__build_replay_context(request_key)
    replay_options = {
      **cli_options,
      **self.__common_replay_options(request_key),
      'mode': mode.MOCK,
    }
    return self.__replay(replay_context, None, replay_options)

  def test(self, request_key: str, cli_options: TestCliOptions):
    replay_context = self.__build_replay_context(request_key)
    replay_options = {
      'mode': mode.TEST,
      'report_key': cli_options.get('report_key'),
      **self.__common_replay_options(request_key),
      **self.common_test_cli_options(cli_options)
    }
    trace_context = replay_options.get('trace_context')

    return self.__replay(replay_context, trace_context, replay_options)

  def __common_replay_options(self, request_key: str):
    return {
      'project_key': ProjectKey.encode(RequestKey(request_key).project_id), # When replaying a specific request, we don't want active project to be used
      'scenario_key': '', # When replaying a specific request, we don't want active scenario to be used
    }

  def __build_replay_context(self, request_key: str):
    request_response, status = self.show(request_key, {
      'body': True,
      'headers': True,
      'query_params': True,
      'response': True,
    })
    return ReplayContext(Request(request_response))

  def __replay(self, replay_context: ReplayContext, trace_context: TraceContext, replay_options: ReplayRequestOptions):
    if not trace_context:
      return replay(replay_context, replay_options)

    group_by = replay_options.get('group_by')
    if not group_by:
      trace_context = replay_options['trace_context']
      return replay_with_trace(replay_context, trace_context, replay_options)
    else:
      return iterate_group_by(
        group_by, 
        trace_context,
        lambda trace_context: replay_with_trace(replay_context.clone(), trace_context, replay_options)
      )


import pdb

from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.lib.api.interfaces import ResponseShowResponse

from ..response import Response as ORMResponse
from ..utils.response_parse_handler import Response as ResponseDict

class ORMToStooblyResponseTransformer():
  __response: ORMResponse = None

  def __init__(self, response: ORMResponse):
    self.__response = response

  @property
  def python_response(self):
    return RawHttpResponseAdapter(self.__response.raw).to_response()

  def transform(self) -> ResponseShowResponse:
    response_dict: ResponseDict = { 'body': self.__response.raw }

    return self.__response_dict_to_response(response_dict)

  def __response_dict_to_response(self, response_dict: ResponseDict) -> ResponseShowResponse:
    stoobly_response: ResponseShowResponse = {}
    stoobly_response['text'] = response_dict['body'] 

    return stoobly_response
  


import click
import collections

from functools import reduce
from typing import List, TypedDict

from stoobly_agent.app.settings import Settings

class CommandGroup(TypedDict):
  commands: List[str]
  name: str

class MainGroup(click.Group):

  def __init__(self, name=None, commands=None, **attrs):
    super(MainGroup, self).__init__(name, commands, **attrs)
    self.commands = commands or collections.OrderedDict()
    self.__settings = Settings.instance()

  def list_commands(self, ctx):
    return self.commands

  def format_commands(self, ctx, formatter):
    super().get_usage(ctx)

    command_groups: List[CommandGroup] = [
      {
        'name': 'Commands',
        'commands': ['dev-tools', 'exec', 'feature', 'init', 'mock', 'record', 'scaffold'],
      },
      {
        'name': 'Proxy Commands',
        'commands': ['ca-cert', 'config', 'intercept', 'run'],
      }
    ]

    if self.__settings.cli.features.remote:
      command_groups.append({
        'name': 'Remote Resource Commands',
        'commands': ['project', 'report', 'request', 'scenario', 'snapshot', 'trace'],
      })
    else:
      command_groups.append({
        'name': 'Local Resource Commands',
        'commands': ['endpoint', 'request', 'scenario', 'snapshot'],
      })

    self.__print(formatter, command_groups)

  def __print(self, formatter, command_groups: List[CommandGroup]):
    command_names = []
    for command_group in command_groups:
      command_names += command_group['commands']

    longest_length_command_length = len(max(command_names, key=len))

    for command_group in command_groups:
      command_names = command_group['commands']
      commands = self.__get_commands(command_names)

      if len(commands) == 0:
        continue

      with formatter.section(command_group['name']):
        for command in commands:
          gap_length = longest_length_command_length - len(command.name) + 2
          formatter.write_text(f"{command.name}{' ' * gap_length}{command.get_short_help_str()}")

  def __get_commands(self, command_names: List[str]):
    return list(filter(
      lambda c: c != None,
      map(lambda c: self.commands.get(c), command_names)
    ))



from .joined_request import REQUEST_DELIMITTER, JoinedRequest
from .request_string import CLRF as REQUEST_STRING_CLRF, RequestString
from .request_string_control import RequestStringControl
from .response_string import ResponseString


import pdb
import random
import re

from time import time

class Cache:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}
            self.dataVersions = {}
            self.startVersion = random.randint(1, 1000000)
            self.timeouts = {}
            self.version = self.startVersion

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    def clear(self, pattern: str = None):
        if not pattern:
            self.data = {}
            self.dataVersions = {}
            self.version += 1
        else:
            delete_list = []
            for key in self.data.keys():
                if re.match(pattern, key):
                    delete_list.append(key)
            
            for key in delete_list:
                self.delete(key)

    def read(self, key):
        if key in self.timeouts:
            timeout = self.timeouts[key]
            if round(time() * 1000) > timeout:
                self.delete(key)
                return None

        data = self.data.get(key)

        if data == None:
            return None

        return {
            'name': key,
            'value': data,
            'version': self.dataVersions[key],
        }

    def read_all(self):
        data = []
        
        for key in list(self.data.keys()):
            datum = self.read(key)

            if datum:
                data.append(datum)

        return data

    # timeout: milliseconds
    def write(self, key, val, timeout = None):
        if timeout and timeout <= 0:
            return

        self.data[key] = val

        if key not in self.dataVersions:
            self.dataVersions[key] = self.startVersion

        if timeout:
            self.timeouts[key] = round(time() * 1000) + timeout

        self.dataVersions[key] += 1
        self.version += 1

    def delete(self, key):
        if key in self.data:
            del self.data[key]

        if key in self.dataVersions:
            del self.dataVersions[key]

        if key in self.timeouts:
            del self.timeouts[key]

        self.version += 1



import jmespath

from typing import Callable, TypedDict, Union

from .visitor import TreeInterpreter, Visitor

# Monkey patch jmespath with replacement functionality
jmespath.parser.visitor.Vistor = Visitor
jmespath.parser.visitor.TreeInterpreter = TreeInterpreter

class Options(TypedDict):
  handle_after_replace: Callable
  handle_replace: Callable

def search(query: str, o: Union[dict, list], options: Options = {}):
  try:
    return jmespath.search(query, o, options)
  except jmespath.exceptions.LexerError as e:
    # On LexerError, try escaping the query
    return jmespath.search(f"\"{query}\"", o, options)

def compile(expression: str):
  return jmespath.compile(expression)

def flatten(value, query: str) -> list:
  array_count = query.count('[*]')

  if array_count == 0 or not isinstance(value, list):
    return [value]
  else:
    return __flatten(value, array_count)

def __flatten(ar, depth, cur_depth = 0):
  if cur_depth == depth:
    return ar

  next_ar = []
  for el in ar:
    if cur_depth == depth - 1:
      next_ar.append(el)
    else:
      if isinstance(el, list):
        next_ar.append(el)
  
  return __flatten(next_ar, depth, cur_depth + 1)


from typing import TypedDict

class RequestColumns(TypedDict):
  body_params_hash: str
  body_text_hash: str
  control: str
  headers_hash: str
  host: str
  http_version: float
  method: str
  path: str
  port: int
  query_params_hash: str
  raw: bytes
  scheme: str


import pdb
import requests

from .python_adapter import PythonRequestAdapter

class MitmproxyRequestAdapterFactory():

  def __init__(self, request: requests.Request):
    self.__request = request

  def python_request(self):
    return PythonRequestAdapter(self.__request).adapt()


import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.cli import record, request
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

@pytest.fixture(autouse=True, scope='module')
def settings():
  return reset()

class TestDelete():

  @pytest.fixture(autouse=True, scope='class')
  def request_create_cli_result(self, runner: CliRunner):
    record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
    assert record_result.exit_code == 0
    return record_result

  @pytest.fixture(autouse=True, scope='class')
  def request_delete_cli_results(self, runner: CliRunner):
    _request = Request.last()
    request_delete_result = runner.invoke(request, ['delete', _request.key()])
    assert request_delete_result.exit_code == 0

    return request_delete_result

  def test_it_sets_is_deleted(self):
    request = Request.last()
    assert request and request.is_deleted

  class TestWhenDeletingAgain():

    def test_it_hard_deletes(self, runner):
      _request = Request.last()
      request_delete_result = runner.invoke(request, ['delete', _request.key()])
      assert request_delete_result.exit_code == 0

      assert Request.last() == None




import pdb
import requests

from mitmproxy.http import Headers, Request as MitmproxyRequest
from urllib.parse import urlparse

from stoobly_agent.lib.utils.decode import decode

class MitmproxyRequestAdapter():

  def __init__(self, http_version: str, request: requests.Request):
    self.__http_version = http_version
    self.__request = request

  @property
  def url(self):
    parsed_url = urlparse(self.__request.url)

    if not parsed_url.netloc:
      parsed_url = parsed_url._replace(netloc=parsed_url.path, path='')

    if not parsed_url.scheme:
      parsed_url = parsed_url._replace(scheme='https')

    return parsed_url.geturl()

  @property
  def headers(self):
    return Headers(**self.__decode_dict(self.__request.headers))

  @property
  def data(self):
    _data = self.__request.data

    # If no data is provided to python requests.Request, it is sent to []
    if _data == []:
      return b''

    return _data

  def adapt(self):
    request = MitmproxyRequest.make(
      self.__request.method,
      self.url,
      self.data,
      self.headers,
    )
    request.http_version = self.__http_version
    return request

  def __decode_dict(self, d):
    new_d = {}
    for k, v in d.items():
      new_d[decode(k)] = decode(v) if isinstance(v, bytes) else v
    return new_d



import pdb

from typing import List

from .firewall_rule import FirewallRule

from .types.proxy_settings import FirewallSettings as IFirewallSettings

class FirewallSettings:
  __firewall_settings = None

  def __init__(self, firewall_settings: IFirewallSettings):
    self.__firewall_settings = firewall_settings or {}

    self.__firewall_rules_map = {
      k: list(map(lambda rule: FirewallRule(rule), v)) for k, v in self.__firewall_settings.items()
    }

  def firewall_rules(self, project_id: str) -> List[FirewallRule]:
    return self.__firewall_rules_map.get(project_id) or []

  def set_firewall_rules(self, project_id: str, v: List[FirewallRule]):
    self.__firewall_rules_map[project_id] = v

  def to_dict(self):
    return {k: list(map(lambda rule: rule.to_dict(), v)) for k, v in self.__firewall_rules_map.items()}


import os
import pdb

from stoobly_agent import VERSION
from stoobly_agent.config.constants.env_vars import ENV
from stoobly_agent.app.settings.constants.mode import TEST

from stoobly_agent.app.settings import Settings
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.orm import ORM
from stoobly_agent.lib.orm.migrate_service import migrate

DETERMINISTIC_GET_REQUEST_URL = 'https://dog.ceo/api/breeds/list/all'
NON_DETERMINISTIC_GET_REQUEST_URL = 'https://www.google.com'

def reset():
  os.environ[ENV] = TEST

  DataDir.instance().remove() # Clean data dir for testing

  ORM.instance().initialize_db()
  migrate(VERSION)

  Settings.instance().reset_and_load()

  return Settings.instance()

def assert_orm_request_equivalent(request_1, request_2):
  assert request_1.latency == request_2.latency
  assert request_1.method == request_2.method
  assert request_1.raw == request_2.raw
  assert request_1.status == request_2.status
  assert request_1.url == request_2.url

reset()



import os
import pdb
import shutil

from click.testing import CliRunner
import pytest

from stoobly_agent.app.cli.scaffold.app import App
from stoobly_agent.app.cli.scaffold.constants import (
  DOCKER_NAMESPACE,
  WORKFLOW_RECORD_TYPE,
)
from stoobly_agent.app.cli.scaffold.service_docker_compose import ServiceDockerCompose
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.test.app.cli.scaffold.e2e_test import ScaffoldCliInvoker
from stoobly_agent.test.test_helper import reset


class TestScaffoldCli():

  @pytest.fixture(scope='module', autouse=True)
  def settings(self):
    return reset()

  @pytest.fixture(scope='module')
  def runner(self):
    yield CliRunner()

  @pytest.fixture(scope='class')
  def app_name(self):
    yield "0.0.1"

  @pytest.fixture(scope='class', autouse=True)
  def temp_dir(self):
    data_dir_path = DataDir.instance().path
    tmp_path = data_dir_path[:data_dir_path.rfind('/')]

    yield tmp_path

  @pytest.fixture(scope='class', autouse=True)
  def app_dir_path(self, temp_dir, app_name):
    yield temp_dir

  @pytest.fixture(scope='class')
  def hostname(self):
    yield "http.badssl.com"

  @pytest.fixture(scope='class')
  def https_service_hostname(self):
    yield "example.com"
  
  @pytest.fixture(scope='class')
  def external_service_name(self):
    yield "external-service"

  @pytest.fixture(scope='class')
  def external_https_service_name(self):
    yield "external-https-service"

  class TestServiceDelete():
    @pytest.fixture(scope='class', autouse=True)
    def target_workflow_name(self):
      yield WORKFLOW_RECORD_TYPE

    @pytest.fixture(scope='class')
    def external_service_docker_compose(self, app_dir_path, target_workflow_name, external_service_name, hostname):
      yield ServiceDockerCompose(app_dir_path=app_dir_path, target_workflow_name=target_workflow_name, service_name=external_service_name, hostname=hostname)
    
    @pytest.fixture(scope='class')
    def external_https_service_docker_compose(self, app_dir_path, target_workflow_name, external_https_service_name, https_service_hostname):
      yield ServiceDockerCompose(app_dir_path=app_dir_path, target_workflow_name=target_workflow_name, service_name=external_https_service_name, hostname=https_service_hostname)

    @pytest.fixture(scope="class", autouse=True)
    def create_setup(self, runner, app_dir_path, app_name, external_service_docker_compose, external_https_service_docker_compose):
      
      ScaffoldCliInvoker.cli_app_create(runner, app_dir_path, app_name)

      # Create external user defined services
      ScaffoldCliInvoker.cli_service_create(runner, app_dir_path, external_service_docker_compose.hostname, external_service_docker_compose.service_name, False)
      ScaffoldCliInvoker.cli_service_create(runner, app_dir_path, external_https_service_docker_compose.hostname, external_https_service_docker_compose.service_name, True)

      # Generate certs
      ScaffoldCliInvoker.cli_app_mkcert(runner, app_dir_path)
    
    @pytest.fixture(scope="class", autouse=True)
    def cleanup_after_all(self, runner, app_dir_path, target_workflow_name):
      yield
      ScaffoldCliInvoker.cli_workflow_down(runner, app_dir_path, target_workflow_name)
      shutil.rmtree(app_dir_path)
    
 
    def test_service_delete(self, runner, app_dir_path, external_service_docker_compose):
      app = App(app_dir_path, DOCKER_NAMESPACE)
      service_name = external_service_docker_compose.service_name
      
      ScaffoldCliInvoker.cli_service_delete(runner, app_dir_path, service_name)

      found = False
      service_paths = app.service_paths
      for service_path in service_paths:
        if service_name in service_path:
          found = True
          break
      if found:
        assert False

    def test_service_delete_https(self, runner, app_dir_path, external_https_service_docker_compose):
      app = App(app_dir_path, DOCKER_NAMESPACE)
      service_name = external_https_service_docker_compose.service_name
      hostname = external_https_service_docker_compose.hostname

      ScaffoldCliInvoker.cli_service_delete(runner, app_dir_path, service_name)

      found = False
      service_paths = app.service_paths
      for service_path in service_paths:
        if service_name in service_path:
          found = True
          break
      if found:
        assert False

      # Certs should be deleted
      certs_dir_path = os.listdir(app.certs_dir_path)
      for cert_file in certs_dir_path:
        if hostname in cert_file:
          assert False




from stoobly_orator.migrations import Migration


class AddOverwritableColumnToScenarios(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.table('scenarios') as table:
            table.boolean('overwritable').default(False)

    def down(self):
        """
        Revert the migrations.
        """
        with self.schema.table('scenarios') as table:
            table.drop_column('overwritable')



import json
import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.cli import record, request
from stoobly_agent.config.constants import test_output_level, test_strategy
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestReplay():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
        return reset()

    class TestWhenDiffStrategy():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        def test_it_matches(self, runner, recorded_request: Request):
            replay_result = runner.invoke(request, ['test', '--strategy', test_strategy.DIFF, recorded_request.key()])
            assert replay_result.exit_code == 0

    class TestWhenContractStrategy():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        def test_it_matches(self, runner, recorded_request: Request):
            replay_result = runner.invoke(request, ['test', '--strategy', test_strategy.CONTRACT, recorded_request.key()])
            assert replay_result.exit_code == 0

    class TestWhenFuzzyStrategy():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        def test_it_matches(self, runner, recorded_request: Request):
            replay_result = runner.invoke(request, ['test', '--strategy', test_strategy.FUZZY, recorded_request.key()])
            assert replay_result.exit_code == 0

    class TestWhenJson():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()
        
        @pytest.fixture(scope='class')
        def tests_output(self, runner: CliRunner, recorded_request: Request):
            replay_result = runner.invoke(request, ['test', '--format', 'json', recorded_request.key()])
            assert replay_result.exit_code == 0
            return json.loads(replay_result.stdout)

        @pytest.fixture(scope='class')
        def test_output(self, tests_output):
            return tests_output['tests'][0]

        def test_it_matches(self, tests_output):
            assert tests_output['passed'] == tests_output['total']

        def test_it_has_no_expected_response(self, test_output):
            assert test_output.get('expected_response') == None

        def test_it_has_log(self, test_output):
            assert test_output.get('log') == ''

        def test_it_has_response(self, test_output):
            assert len(test_output.get('response').keys() - ['content', 'status_code', 'latency']) == 0

    class TestWhenFailedOutputLevel():

        @pytest.fixture(scope='class')
        def recorded_request(self, runner: CliRunner):
            record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()
        
        @pytest.fixture(scope='class')
        def test_output(self, runner: CliRunner, recorded_request: Request):
            replay_result = runner.invoke(request, ['test', '--output-level', test_output_level.FAILED, recorded_request.key()])
            assert replay_result.exit_code == 0
            return replay_result.stdout

        def test_it_has_less_output(self, runner: CliRunner, test_output: str, recorded_request: Request):
            replay_result = runner.invoke(request, ['test', recorded_request.key()])
            assert replay_result.exit_code == 0
            assert len(test_output) < len(replay_result.stdout) 

        class TestWhenJson():
            @pytest.fixture(scope='class')
            def recorded_request(self, runner: CliRunner):
                record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
                assert record_result.exit_code == 0
                return Request.last()
            
            @pytest.fixture(scope='class')
            def tests_output(self, runner, recorded_request: Request):
                replay_result = runner.invoke(request, ['test', '--output-level', test_output_level.FAILED, '--format', 'json', recorded_request.key()])
                assert replay_result.exit_code == 0
                return json.loads(replay_result.stdout)

            @pytest.fixture(scope='class')
            def test_output(self, tests_output):
                return tests_output['tests'][0]

            def test_it_has_no_expected_response(self, test_output):
                assert test_output.get('expected_response') == None

            def test_it_has_no_log(self, test_output):
                assert test_output.get('log') == None

            def test_it_has_no_response(self, test_output):
                assert test_output.get('response') == None


import os
import pdb

from typing import List

from stoobly_agent.config.data_dir import DATA_DIR_NAME

from ...app_config import AppConfig
from ...constants import (
  APP_DIR, DOCKER_NAMESPACE, SERVICE_NAME, SERVICE_NAME_ENV, 
  SERVICE_HOSTNAME, SERVICE_HOSTNAME_ENV, SERVICE_NAME_ENV, SERVICE_PORT, SERVICE_PORT_ENV, SERVICE_SCHEME, SERVICE_SCHEME_ENV, 
  STOOBLY_HOME_DIR, STOOBLY_HOME_DIR, 
  WORKFLOW_NAME, WORKFLOW_NAME_ENV, WORKFLOW_SCRIPTS, WORKFLOW_TEMPLATE
)
from ...service_config import ServiceConfig
from ..app_builder import AppBuilder
from ..builder import Builder
from ..constants import DOCKER_COMPOSE_BASE

class ServiceBuilder(Builder):

  def __init__(self, config: ServiceConfig, app_builder: AppBuilder = None):
    service_path = config.dir
    super().__init__(service_path, DOCKER_COMPOSE_BASE)

    if not app_builder:
      app_dir = os.path.dirname(service_path)
      app_builder = AppBuilder(AppConfig(app_dir))
    self.app_builder = app_builder

    self.__config = config
    self.__env = [SERVICE_NAME_ENV, WORKFLOW_NAME_ENV]
    self.__service_name = os.path.basename(service_path)
    self.__working_dir = os.path.join(
      STOOBLY_HOME_DIR, DATA_DIR_NAME, DOCKER_NAMESPACE, SERVICE_NAME, WORKFLOW_NAME
    )

  @property
  def app_base(self):
    return f"{self.service_name}.app_base"

  @property
  def config(self):
    return self.__config

  @property
  def configure_base(self):
    return f"{self.service_name}.configure_base"

  @property
  def configure_base_service(self):
    return self.services.get(self.configure_base)

  @property
  def extends_service(self):
    if self.config.detached:
      return self.app_builder.stoobly_base
    else:
      return self.app_builder.context_base

  @property
  def init_base(self):
    return f"{self.service_name}.init_base"

  @property
  def init_base_service(self):
    return self.services.get(self.init_base)

  @property
  def proxy_base(self):
    return f"{self.service_name}.proxy_base"

  @property
  def proxy_base_service(self):
    return self.services.get(self.proxy_base)

  @property
  def service_name(self):
    return self.__service_name

  def build_extends_init_base(self, source_dir: str):
    return self.build_extends(self.init_base, source_dir)

  def build_extends_configure_base(self, source_dir: str):
    return self.build_extends(self.configure_base, source_dir)

  def build_extends_proxy_base(self, source_dir: str):
    return self.build_extends(self.proxy_base, source_dir)

  def build_proxy_base(self):
    if not self.config.hostname:
      return

    service_id = self.config.id
    environment = { **self.env_dict() }
    labels = [
      'traefik.enable=true',
      f"traefik.http.routers.{service_id}.rule=Host(`{SERVICE_HOSTNAME}`)",
      f"traefik.http.routers.{service_id}.entrypoints={SERVICE_PORT}",
      f"traefik.http.services.{service_id}.loadbalancer.server.port={SERVICE_PORT}"
    ]
    volumes = []

    if self.config.detached:
      self.__with_detached_volumes(volumes)

    if self.config.tls:
      labels.append(f"traefik.http.routers.{service_id}.tls=true")

    base = {
      'environment': environment,
      'extends': {
        'file': os.path.relpath(self.app_builder.compose_file_path, self.dir_path),
        'service': self.extends_service
      },
      'labels': labels,
      'working_dir': self.__working_dir,
    }

    if len(volumes):
      base['volumes'] = volumes

    self.with_service(self.proxy_base, base)

  def build_init_base(self):
    environment = { **self.env_dict() }
    volumes = [f"{APP_DIR}:/app"]

    if self.config.hostname:
      self.__with_url_environment(environment)

    if self.config.detached:
      self.__with_detached_volumes(volumes)

    self.with_service(self.init_base, {
      'command': [f"{WORKFLOW_SCRIPTS}/{WORKFLOW_TEMPLATE}/.init", 'bin/init'],
      'environment': environment,
      'extends': {
        'file': os.path.relpath(self.app_builder.compose_file_path, self.dir_path),
        'service': self.extends_service
      },
      'volumes': volumes,
      'working_dir': self.__working_dir,
    })

  def build_configure_base(self):
    environment = { **self.env_dict() }
    volumes = []

    if self.config.hostname:
      self.__with_url_environment(environment)

    if self.config.detached:
      self.__with_detached_volumes(volumes)

    base = {
      'command': [f"{WORKFLOW_SCRIPTS}/{WORKFLOW_TEMPLATE}/.configure", 'bin/configure'],
      'environment': environment,
      'extends': {
        'file': os.path.relpath(self.app_builder.compose_file_path, self.dir_path),
        'service': self.extends_service
      },
      'working_dir': self.__working_dir,
    }

    if len(volumes):
      base['volumes'] = volumes

    self.with_service(self.configure_base, base)

  def env_dict(self):
    env = {}
    for e in self.__env:
      env[e] = '${' + e + '}'
    return env

  def with_env(self, v: List[str]): 
    if not isinstance(v, list):
      return self
    self.__env += v
    return self

  def write(self):
    self.build_init_base()
    self.build_configure_base()

    if self.config.hostname:
      self.build_proxy_base()

    compose = {
      'services': self.services,
    }

    if self.networks:
      compose['networks'] = self.networks

    super().write(compose)

  def __with_detached_volumes(self, volumes: list):
    # Mount named volume
    volumes.append(f"{self.service_name}:{STOOBLY_HOME_DIR}/{DATA_DIR_NAME}")

    # Mount docker folder
    volumes.append(f"../:{STOOBLY_HOME_DIR}/{DATA_DIR_NAME}/{DOCKER_NAMESPACE}")

  def __with_url_environment(self, environment):
    environment[SERVICE_HOSTNAME_ENV] = SERVICE_HOSTNAME

    if self.config.scheme:
      environment[SERVICE_SCHEME_ENV] = SERVICE_SCHEME

    if self.config.port:
      environment[SERVICE_PORT_ENV] = SERVICE_PORT


import pdb

from urllib.parse import urlparse

from stoobly_agent.lib.api.keys.request_key import RequestKey
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey

def search_request(base_model, query: str):
  try:
    key = RequestKey(query)
    return base_model.where('uuid', key.id)
  except Exception as e:
    uri = urlparse(query)

    if uri.hostname:
      return base_model.where('host', uri.hostname).where('path', uri.path)
    else:
      pattern = f"%{query}%"
      return base_model.where('path', 'like', pattern).or_where('host', 'like', pattern)

def search_scenario(base_model, query: str):
  try:
    key = ScenarioKey(query)
    return base_model.where('uuid', key.id)
  except Exception as e:
    return base_model.where('name', 'like', f"%{query}%")


import base64
import json
import pdb
import requests

from typing import List, Tuple

from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter
from stoobly_agent.lib.api.interfaces import Header, HeaderShowResponse
from stoobly_agent.lib.orm.request import Request

from .local_db_adapter import LocalDBAdapter
from .request_adapter import LocalDBRequestAdapter

class LocalDBHeaderAdapter(LocalDBAdapter):
  __request_orm = None

  def __init__(self, request_orm: Request.__class__ = Request):
    self.__request_orm = request_orm

  def create(self, request_id, **header: Header):
    return self.update(request_id, **header)

  def update(self, request_id, **header: Header):
    request = self.__request_orm.find(request_id)
    
    if not request:
      return self.__request_not_found()

    name = header['name']
    value = header['value']

    request: requests.Request = RawHttpRequestAdapter(request.raw).to_request()
    headers = request.headers
    headers[name] = value

    LocalDBRequestAdapter(self.__request_orm).update(request_id, headers=headers)

    return self.success({
      'name': name,
      'value': value,
    })

  def index(self, request_id, **query_params) -> Tuple[List[HeaderShowResponse], int]:
    request = self.__request_orm.find(request_id)

    if not request:
      return self.__request_not_found()

    request: requests.Request = RawHttpRequestAdapter(request.raw).to_request()

    headers = []
    for key, val in request.headers.items():
      headers.append({
        'name': key,
        'value': val,
      })

    return self.success(headers)

  def destroy(self, request_id, id):
    request = self.__request_orm.find(request_id)
    
    if not request:
      return self.__request_not_found()

    decoded_id = self.__decode_id(id)
    if not decoded_id:
      return self.bad_request('Invalid id')

    python_request: requests.Request = RawHttpRequestAdapter(request.raw).to_request()
    headers = python_request.headers

    header_name = decoded_id['name']
    if not header_name in headers:
      return self.not_found()

    del headers[header_name]

    LocalDBRequestAdapter(self.__request_orm).update(request_id, headers=headers)

    return self.success({
      'name': decoded_id['name'],
    })

  def __decode_id(self, id: str):
    id = base64.b64decode(id)

    try:
      id = json.loads(id)
    except Exception as e:
      return None

    if not isinstance(id, dict):
      return None

    if not id.get('name'):
      return None

    return id

  def __request_not_found(self):
    return self.not_found('Request not found')


import base64
import copy
import pdb

from mitmproxy.net.http.url import encode as urlencode
from mitmproxy.coretypes.multidict import MultiDict
from urllib.parse import urlparse
from typing import List, Union

from stoobly_agent.app.proxy.replay.body_parser_service import decode_response, encode_response, is_traversable
from stoobly_agent.lib.api.interfaces.endpoints import RequestComponentName
from stoobly_agent.lib.api.interfaces import RequestShowResponse

class Request():

  def __init__(self, request: RequestShowResponse):
    self.request = request
    self.__url = urlparse(request.get('url'))

  def clone(self):
    return Request(copy.deepcopy(self.request))

  @property
  def host(self):
    return self.__url.netloc

  @host.setter
  def host(self, v):
    self.__url = self.__url._replace(netloc=v)

  @property
  def http_version(self):
    return self.request.get('http_version')

  @property
  def key(self):
    return self.request.get('key')

  @property
  def scheme(self):
    return self.__url.scheme

  @scheme.setter
  def scheme(self, v):
    self.__url = self.__url._replace(scheme=v)

  @property
  def id(self):
    return self.request.get('id')

  @property
  def endpoint_id(self):
    return self.request.get('endpoint_id')

  @property
  def url(self) -> str:
    query_params = self.query_params

    query_tuples = []
    for key in query_params:
      for value in query_params.get_all(key):
        if not isinstance(value, list):
          query_tuples.append((key, value))
        else:
          for v in value:
            query_tuples.append((key, v))

    url = self.__url._replace(path=self.path, query=urlencode(query_tuples))
    return url.geturl()

  @property
  def path(self) -> str:
    return self.request.get('path') or ''

  @path.setter
  def path(self, v):
    self.request['path'] = v

  @property
  def path_segment_strings(self):
    segment_strings = self.path.split('/')
    if len(segment_strings) == 0:
      return segment_strings

    if segment_strings[0] == '':
      segment_strings.pop(0)

    return segment_strings

  @property
  def method(self) -> str:
    return self.request['method']

  @property
  def headers(self) -> MultiDict:
    headers = MultiDict()

    for header in (self.request.get('headers') or []):
      headers.add(header['name'], header['value'])

    return headers 

  @headers.setter
  def headers(self, v: MultiDict):
    self.__set_dict('headers', v)

  @property
  def query(self) -> str:
    return urlencode(self.query_params)

  @property
  def query_params(self) -> MultiDict:
    query_params = MultiDict()

    for query_param in (self.request.get('query_params') or []):
      query_params.add(query_param['name'], query_param['value'])

    return query_params

  @query_params.setter
  def query_params(self, v: MultiDict):
    self.__set_dict('query_params', v) 
  
  @property
  def body(self):
    encoded_body = self.request.get('body')
    if not encoded_body:
      return ''

    try:
      return base64.b64decode(encoded_body)
    except Exception as e:
      return encoded_body

  @body.setter
  def body(self, v: Union[bytes, str]):
    if not 'body' in self.request:
      return

    text = v if isinstance(v, bytes) else v.encode()
    self.request['body'] = base64.b64encode(text)

    headers = self.headers
    if 'content-length' in headers:
      headers['content-length'] = str(len(text))
      self.headers = headers

  @property
  def body_params(self):
    _body = self.body
    return decode_response(_body, self.headers.get('content-type'))

  @body_params.setter
  def body_params(self, v):
    if not is_traversable(v):
      self.body = v
    else:
      content_type = self.headers.get('content-type')
      if content_type:
        self.body = encode_response(v, content_type)

  def __set_dict(self, component_name_str, v: MultiDict):
    component_names: List[RequestComponentName] = self.request[component_name_str] or []

    indexes = {}
    new_component_names = []
    for component_name in component_names:
      name = component_name['name']
      if name in v:
        value = v.get_all(name)

        if isinstance(value, list):
          if name not in indexes:
            indexes[name] = 0
          else:
            indexes[name] += 1

          index = indexes[name]

          if len(value) <= index:
            continue

          component_name['value'] = value[index]
        else:
          component_name['value'] = value

      new_component_names.append(component_name)

    # If a component was set that doesn't exist in the request template, add it
    for key in v:
      if key in indexes:
        continue
      
      value = v.get_all(key)

      if isinstance(value, list):
        for val in value:
          new_component_names.append({
            'name': key,
            'value': val,
          })
      else:
        new_component_names.append({
          'name': key,
          'value': value,
        })

    self.request[component_name_str] = new_component_names


import requests

from mitmproxy.http import Request as MitmproxyRequest

class PythonRequestAdapter():
  
  def __init__(self, request: MitmproxyRequest):
    self.__request = request

  def adapt(self):
    return requests.Request(
      method=self.__request.method,
      url=self.__request.url,
      headers=self.__request.headers,
      data=self.__request.content
    )



import pdb
import re
from time import sleep
from typing import Union

import docker
from docker import errors as docker_errors
from docker.models.containers import Container

from stoobly_agent.app.cli.scaffold.validate_exceptions import ScaffoldValidateException
from stoobly_agent.config.data_dir import DATA_DIR_NAME
from stoobly_agent.lib.logger import bcolors


class ValidateCommand():
  def __init__(self):
    self.docker_client = docker.from_env()

  def __generate_container_not_found_error(self, container_name: Union[str, None]) -> str:
    not_found_error_message = f"{bcolors.FAIL}Container not found: {container_name}{bcolors.ENDC}"
    suggestion_message = f"{bcolors.BOLD}Run 'docker ps -a | grep {container_name}'. If found, then inspect the logs with 'docker logs {container_name}'{bcolors.ENDC}"
    error_message = f"{not_found_error_message}\n\n{suggestion_message}"

    return error_message

  # Some containers like init and configure can take longer than expected to finish so retry
  def __get_container_with_retries(self, container_name: str) -> Container:
    tries = 30
    for _ in range(tries):
      try:
        container = self.docker_client.containers.get(container_name)
        return container
      except docker_errors.NotFound:
        sleep(0.5)

    error_message = self.__generate_container_not_found_error(container_name)
    raise ScaffoldValidateException(error_message)

  def validate_init_containers(self, init_container_name, configure_container_name) -> None:
    print(f"Validating setup containers: {init_container_name}, {configure_container_name}")

    init_container = self.__get_container_with_retries(init_container_name)
    logs = init_container.logs()

    if logs and re.search('error', str(logs), re.IGNORECASE):
      error_found_message = f"{bcolors.FAIL}Error logs potentially detected in: {init_container_name}{bcolors.ENDC}"
      suggestion_message = f"{bcolors.BOLD}Run 'docker logs {init_container_name}'{bcolors.ENDC}"
      error_message = f"{error_found_message}\n\n{suggestion_message}"
      raise ScaffoldValidateException(error_message)

    if init_container.status != 'exited' or init_container.attrs['State']['ExitCode'] != 0:
      init_exit_message = f"{bcolors.FAIL}init container {init_container_name} exited with: {init_container.attrs['State']['ExitCode']}{bcolors.ENDC}"
      suggestion_message = f"{bcolors.BOLD}Run 'docker logs {init_container_name}'{bcolors.ENDC}"
      error_message = f"{init_exit_message}\n\n{suggestion_message}"
      raise ScaffoldValidateException(error_message)

    configure_container = self.__get_container_with_retries(configure_container_name)

    configure_container_ran = False
    if configure_container.status == 'exited' and configure_container.attrs['State']['ExitCode'] == 0:
      configure_container_ran = True
    if not configure_container_ran:
      raise ScaffoldValidateException(f"configure container {configure_container_name} exited with: {configure_container.attrs['State']['ExitCode']}")
  
  def validate_detached(self, container: Container) -> None:
    print(f"Validating detached for: {container.name}")
    
    if not container.attrs:
      error_message = self.__generate_container_not_found_error(container.name)
      raise ScaffoldValidateException(error_message)

    volume_mounts = container.attrs['Mounts']
    for volume_mount in volume_mounts:
      if DATA_DIR_NAME in volume_mount['Source']:
        return
    
    message = f"{bcolors.FAIL}Data directory is missing from container: {container.name}{bcolors.ENDC}"
    suggestion_message = f"{bcolors.BOLD}Data directory might have failed to mount{bcolors.ENDC}"
    error_message = f"{message}\n\n{suggestion_message}"

    raise ScaffoldValidateException(error_message)




import pdb

from typing import Union

from stoobly_agent.app.proxy.record.joined_request import JoinedRequest, REQUEST_DELIMITTER
from stoobly_agent.app.proxy.record.request_string import CLRF as RequestStringCLRF, RequestString
from stoobly_agent.app.proxy.record.response_string import CLRF as ResponseStringCLRF, ResponseString

class JoinedRequestAdapter():

  def __init__(self, joined_request_string: Union[bytes, str], payloads_delimitter: Union[bytes, str] = None):
    payloads_delimitter = REQUEST_DELIMITTER if not payloads_delimitter else payloads_delimitter

    if isinstance(joined_request_string, str):
      joined_request_string = joined_request_string.encode()

    if isinstance(payloads_delimitter, str):
      payloads_delimitter = payloads_delimitter.encode()
      
    self.__split_joined_request_string = joined_request_string.split(payloads_delimitter)
    if len(self.__split_joined_request_string) != 2:
      self.__split_joined_request_string = joined_request_string.split(payloads_delimitter.replace(b"\n", b"\r\n"))

    if len(self.__split_joined_request_string) != 2:
      raise ValueError(f"Could not split by {payloads_delimitter}")

    self.__raw_request_string = None
    self.__raw_response_string = None

  @property
  def raw_request_string(self):
    return self.__raw_request_string

  @raw_request_string.setter
  def raw_request_string(self, v):
    self.__raw_request_string = v

  @property
  def raw_response_string(self):
    return self.__raw_response_string

  @raw_response_string.setter
  def raw_response_string(self, v):
    self.__raw_response_string = v

  def build_request_string(self):
    request_string = RequestString(None)

    delimitter = RequestStringCLRF.encode()
    request_string_toks = self.__split_joined_request_string[0].split(delimitter)
    request_string.set(self.raw_request_string or delimitter.join(request_string_toks[1:]))
    request_string.control = request_string_toks[0]

    return request_string

  def build_response_string(self):
    response_string = ResponseString(None, None)

    delimitter = ResponseStringCLRF
    response_string_toks = self.__split_joined_request_string[1].split(delimitter)
    response_string.set(self.raw_response_string or delimitter.join(response_string_toks[1:]))
    response_string.control = response_string_toks[0]

    return response_string

  def adapt(self) -> JoinedRequest:
    joined_request = JoinedRequest()

    joined_request.request_string = self.build_request_string()
    joined_request.response_string = self.build_response_string()
    return joined_request


import json
import pdb
import requests
import sys

from typing import TypedDict, Union

from stoobly_agent.app.cli.helpers.handle_replay_service import JSON_FORMAT, format_request, print_with_decoding
from stoobly_agent.app.cli.helpers.test_facade import TestFacade
from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.cli.types.output import TestOutput
from stoobly_agent.config.constants import test_output_level
from stoobly_agent.lib.api.interfaces.tests import TestShowResponse
from stoobly_agent.lib.logger import bcolors
from stoobly_agent.lib.utils.decode import decode

from .test_replay_context import TestReplayContext

class HandleOptions(TypedDict):
  format: str

class HandleTestOptions(HandleOptions):
  output_level: test_output_level.TestOutputLevel

class SessionContext(TypedDict):
  aggregate_failures: bool 
  passed: int
  output: TestOutput
  project_id: int
  skipped: int
  test_facade: TestFacade
  total: int

class ExitOnFailureOptions(TypedDict):
  complete: bool
  format: Union[str, None] 

def handle_test_session_complete(session_context: SessionContext, **options: HandleOptions):
  format = options.get('format')
  format_handler = __default_session_complete_formatter

  if format == JSON_FORMAT:
    format_handler = __json_session_complete_formatter

  format_handler(session_context)

def handle_test_complete(
  context: ReplayContext, session_context: SessionContext, **options: HandleTestOptions 
):
  format = options.get('format')
  format_handler = __default_test_complete_formatter

  if format == JSON_FORMAT:
    format_handler = __json_test_complete_formatter

  # If not context.has_test_results, then test results are from remote
  # Configure which project test the test results are from
  context = TestReplayContext(context)
  if not context.has_test_results:
    context.project_id = session_context['project_id']

  test = context.test_show_response(session_context['test_facade'])

  passed = bool(test.get('passed')) if isinstance(test, dict) else False
  skipped = bool(test.get('skipped')) if isinstance(test, dict) else False

  session_context['skipped'] += (1 if skipped else 0)
  session_context['passed'] += (1 if passed else 0)
  session_context['total'] += 1

  format_handler(context, session_context, test, **options)

def exit_on_failure(session_context: SessionContext, **options: ExitOnFailureOptions):
  complete = options.get('complete') 
  if complete == None:
    complete = True

  if session_context['passed'] + session_context['skipped'] != session_context['total']:
    if not complete:
      handle_test_session_complete(session_context, format=options.get('format'))

    sys.exit(1)

def __should_output(output_level: test_output_level.TestOutputLevel, test: TestShowResponse):
  if not output_level:
    return True

  if output_level == test_output_level.PASSED:
    return True

  passed = bool(test.get('passed')) if isinstance(test, dict) else False
  skipped = bool(test.get('skipped')) if isinstance(test, dict) else False

  output = True 
  if output_level == test_output_level.FAILED:
    if passed or skipped:
      output = False
  elif output_level == test_output_level.SKIPPED:
    if passed:
      output = False

  return output

def __json_test_complete_formatter(
  context: TestReplayContext, session_context: SessionContext, res: TestShowResponse, **options: HandleTestOptions
):
  if 'output' not in session_context:
    session_context['output'] = {}

  output = session_context['output']
  if not res:
    output['error'] = 'Test failed to run'
  elif isinstance(res, str):
    output['error'] = res
  else:
    if 'tests' not in output:
      output['tests'] = []

    request = context.request
    seconds = context.end_time - context.start_time
    latency = round(seconds * 1000)
    test = {
      'key': context.key,
      'method': request.method,
      'passed': res['passed'],
      'url': request.url,
    }

    if __should_output(options.get('output_level'), res):
      test['log'] = res['log'] if 'log' in res else ''
      test['response'] = { **__build_json_response(context.response), 'latency': latency}

      if not res['passed']:
        test_facade = session_context['test_facade']
        expected_response, expected_status_code = context.expected_response_content(test_facade)

        if expected_response:
          test['expected_response'] = { 
            'content': expected_response,
            'latency': res['expected_latency'],
            'status_code': expected_status_code,
          }

    output['tests'].append(test)

def __default_test_complete_formatter(
  context: TestReplayContext, session_context: SessionContext, res: TestShowResponse, **options: HandleTestOptions
):
  if not res:
    print_with_decoding(format_request(context))
    print("\nTest failed to run")
  elif isinstance(res, str):
    print(res, file=sys.stderr)
  else:
    should_output = __should_output(options.get('output_level'), res)

    if should_output:
      print_with_decoding(format_request(context))

    if not res['passed']:
      if res['skipped']:
        passed_message = f"{bcolors.WARNING}skipped{bcolors.ENDC}"
      else:
        passed_message = f"{bcolors.FAIL}failed{bcolors.ENDC}"
    else:
      passed_message = f"{bcolors.OKGREEN}passed{bcolors.ENDC}"

    print(f"\nTest {passed_message} using {bcolors.BOLD}{res['strategy']}{bcolors.ENDC} strategy")
    
    if should_output:
      if res['log']:
        print(f"\n{res['log']}")

      if not res['passed'] and not res['skipped']:
        test_facade = session_context['test_facade']
        response, status_code, latency = context.diff_response(res, test_facade)
        print("\n" + response) 
        print(f"Completed {status_code} in {latency}ms")

def __default_session_complete_formatter(session_context: SessionContext):
  if 'output' in session_context:
    print(session_context['output'])

  total = session_context['total']

  if total > 1:
    passed = session_context['passed']
    skipped = session_context['skipped']

    passed_message = f"{passed} / {total} {bcolors.OKGREEN}passed{bcolors.ENDC}"

    skipped_message = ''
    if skipped:
      skipped_message = f" {skipped} {bcolors.WARNING}skipped{bcolors.ENDC}"

    if passed == total:
      print(f"\n{passed_message}{skipped_message}")
    else:
      failed_message = f"{total - passed - skipped} {bcolors.FAIL}failed{bcolors.ENDC}"
      print(f"\n{passed_message} {failed_message}{skipped_message}")

def __json_session_complete_formatter(session_context: SessionContext):
  if 'output' not in session_context:
    return

  session_context['output']['passed'] = session_context['passed']
  session_context['output']['skipped'] = session_context['skipped']
  session_context['output']['total'] = session_context['total']

  print(json.dumps(session_context['output']))

def __build_json_response(response: requests.Response):
  return {
    'content': __decode_response(response),
    'status_code': response.status_code,
  }

def __decode_response(response: requests.Response):
  if not isinstance(response, requests.Response):
    return ''
  content = response.content
  return decode(content)


import pdb
import pytest
import time

from click.testing import CliRunner

from stoobly_agent.app.models.factories.resource.local_db.helpers.log import Log
from stoobly_agent.app.models.factories.resource.local_db.helpers.log_event import DELETE_ACTION, LogEvent
from stoobly_agent.cli import record, request, snapshot
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestPrune():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
      return reset()

    class TestWhenDeletingRequest():

      @pytest.fixture(scope='class')
      def log(self):
        return Log()

      @pytest.fixture(scope='class')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      @pytest.fixture(autouse=True, scope='class')
      def events(self, runner: CliRunner, recorded_request: Request, log: Log):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
        assert snapshot_result.exit_code == 0

        events = log.events
        event: LogEvent = events[len(events) - 1]
        log.version = event.uuid

        time.sleep(1)

        snapshot_result = runner.invoke(request, ['snapshot', '--action', DELETE_ACTION, recorded_request.key()])
        assert snapshot_result.exit_code == 0

        return log.events

      def test_it_prunes(self, runner: CliRunner, log: Log):
        assert len(log.events) == 2
        apply_result = runner.invoke(snapshot, ['prune'])
        assert apply_result.exit_code == 0
        assert len(log.events) == 0

    class TestWhenCreatingRequest():

      @pytest.fixture(scope='class')
      def log(self):
        return Log()

      @pytest.fixture(scope='function')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      @pytest.fixture(autouse=True, scope='function')
      def put_event(self, runner: CliRunner, recorded_request: Request):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
        assert snapshot_result.exit_code == 0

        log = Log()
        events = log.events
        return events[len(events) - 1]

      def test_it_does_not_prune(self, runner: CliRunner, log: Log):
        apply_result = runner.invoke(snapshot, ['prune'])
        assert apply_result.exit_code == 0
        assert len(log.events) == 1

    class TestWhenMissingSnapshot():

      @pytest.fixture(scope='class')
      def log(self):
        return Log()

      @pytest.fixture(scope='function')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      @pytest.fixture(autouse=True, scope='function')
      def put_event(self, runner: CliRunner, recorded_request: Request):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
        assert snapshot_result.exit_code == 0

        log = Log()
        events = log.events
        return events[len(events) - 1]

      def test_it_prunes(self, runner: CliRunner, log: Log):
        event = log.events[0]
        event.snapshot().remove()
        assert len(log.events) == 1

        apply_result = runner.invoke(snapshot, ['prune'])
        assert apply_result.exit_code == 0
        assert len(log.events) == 0


import pdb
import requests

from mitmproxy.http import Response as MitmproxyResponse

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, NON_DETERMINISTIC_GET_REQUEST_URL

from stoobly_agent.app.models.adapters.python import PythonResponseAdapterFactory

class TestMitmproxyResponseAdapter():

  def test_deterministic_request(self):
    res = requests.get(DETERMINISTIC_GET_REQUEST_URL) 

    mitmproxy_response = PythonResponseAdapterFactory(res).mitmproxy_response()
    self.__test_equivalence(mitmproxy_response, res)

  def test_non_deterministic_request(self):
    res = requests.get(NON_DETERMINISTIC_GET_REQUEST_URL) 

    mitmproxy_response = PythonResponseAdapterFactory(res).mitmproxy_response()
    self.__test_equivalence(mitmproxy_response, res)

  def __test_equivalence(self, mitmproxy_response: MitmproxyResponse, response: requests.Response):
    assert mitmproxy_response.raw_content == response.raw.data
    assert mitmproxy_response.content == response.content
    assert mitmproxy_response.status_code == response.status_code

    for key, val in response.headers.items():
      assert val == mitmproxy_response.headers.get(key)




import pdb

from typing import TypedDict

from stoobly_agent.app.models.scenario_model import ScenarioModel
from stoobly_agent.app.settings.constants import intercept_mode
from stoobly_agent.app.settings.constants.intercept_mode import Mode
from stoobly_agent.app.settings import ProxySettings, Settings
from stoobly_agent.config.constants import record_policy
from stoobly_agent.lib.api.keys.project_key import ProjectKey
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey

class Context(TypedDict):
  current_proxy_settings: ProxySettings
  mode: Mode

def context() -> Context:
  return {
    'current_proxy_settings': __current_proxy_settings()
  }

def handle_intercept_active_update(new_settings: Settings, context: Context = None):
  data_rule = __data_rule(new_settings.proxy)
  is_active = new_settings.proxy.intercept.active
  old_proxy_settings = __current_proxy_settings(context)
  was_active = old_proxy_settings.intercept.active
  _mode = context['mode'] if context and context.get('mode') else new_settings.proxy.intercept.mode

  if not was_active and is_active:
    if _mode == intercept_mode.RECORD:
      new_policy = data_rule.record_policy
      scenario_key = data_rule.scenario_key
      _scenario_key = __parse_scenario_key(scenario_key)

      if _scenario_key:
        scenario_model = ScenarioModel(new_settings)

        if new_policy == record_policy.OVERWRITE:
          # If policy is overwrite when recording, whenever intercept is enabled,
          # set active scenario to be overwritable
          scenario_model.update(_scenario_key.id, **{ 'overwritable': True })[1]
        else:
          scenario_model.update(_scenario_key.id, **{ 'overwritable': False })[1]
  elif was_active and not is_active:
    if _mode == intercept_mode.RECORD:
      old_proxy_settings = __current_proxy_settings(context)
      old_policy = __data_rule(old_proxy_settings).record_policy

      if old_policy == record_policy.OVERWRITE:
        scenario_key = data_rule.scenario_key
        _scenario_key = __parse_scenario_key(scenario_key)

        if _scenario_key:
          scenario_model = ScenarioModel(new_settings)

          # If policy is overwrite when recording, whenever intercept is disabled,
          # set active scenario to not be overwritable
          scenario_model.update(_scenario_key.id, **{ 'overwritable': False })[1]
    elif _mode == intercept_mode.MOCK:
      # When mock is stopped, clear request access counts
      from stoobly_agent.app.models.factories.resource.local_db.helpers.tiebreak_scenario_request import reset_sessions

      reset_sessions()

def handle_scenario_update(new_settings: Settings, context = None):
  new_scenario_key = __scenario_key(new_settings.proxy)
  if not new_scenario_key:
    return

  old_proxy_settings = __current_proxy_settings(context)
  old_scenario_key = __scenario_key(old_proxy_settings)

  if old_scenario_key != new_scenario_key:
    data_rule = __data_rule(new_settings.proxy)

    if data_rule.record_policy == record_policy.OVERWRITE:
      scenario_model = ScenarioModel(new_settings)

      _old_scenario_key = __parse_scenario_key(old_scenario_key)
      if _old_scenario_key:
        scenario_model.update(_old_scenario_key.id, **{ 'overwritable': False })[1]

      _new_scenario_key = __parse_scenario_key(new_scenario_key)
      if new_scenario_key:
        scenario_model.update(_new_scenario_key.id, **{ 'overwritable': True })[1]

def handle_project_update(new_settings: Settings, context: Context = None):
  new_project_key = new_settings.proxy.intercept.project_key
  if not new_project_key:
    return

  old_proxy_settings = __current_proxy_settings(context)
  old_project_key = __project_key(old_proxy_settings)

  # If the active project changed, stop intercepting 
  if new_project_key != old_project_key:
    if old_proxy_settings.intercept.active:
        new_settings.proxy.intercept.active = False

  # Check to see if scenario belongs to current project
  data_rule = __data_rule(old_proxy_settings)
  new_scenario_key = data_rule.scenario_key

  _new_scenario_key = __parse_scenario_key(new_scenario_key)
  if _new_scenario_key:
    _project_key = ProjectKey(new_project_key)
    
    if _project_key.id != _new_scenario_key.project_id:
      data_rule.scenario_key = None

def handle_policy_update(new_settings: Settings, context: Context = None):
  data_rule = __data_rule(new_settings.proxy)
  _mode = context['mode'] if context and context.get('mode') else new_settings.proxy.intercept.mode

  old_proxy_settings = __current_proxy_settings(context)
  old_data_rule = __data_rule(old_proxy_settings)

  if _mode == intercept_mode.RECORD:
    new_policy = data_rule.record_policy
    old_policy = old_data_rule.record_policy

    if new_policy != old_policy and old_policy == record_policy.OVERWRITE:
      scenario_key = data_rule.scenario_key
      _scenario_key = __parse_scenario_key(scenario_key)

      if _scenario_key:
        scenario_model = ScenarioModel(new_settings)
        scenario_model.update(_scenario_key.id, **{ 'overwritable': False })[1]

def __current_proxy_settings(context: Context = None):
  if context and context.get('current_proxy_settings'):
    return context['current_proxy_settings']

  old_proxy_settings_raw = Settings.instance().read().get('proxy')
  return ProxySettings(old_proxy_settings_raw)

def __data_rule(proxy_settings: ProxySettings):
  project_key = proxy_settings.intercept.project_key
  
  _project_key = ProjectKey(project_key)
  return proxy_settings.data.data_rules(_project_key.id)

def __parse_scenario_key(scenario_key: str):
  try:
    return ScenarioKey(scenario_key)
  except Exception as e:
    return

def __project_key(proxy_settings: ProxySettings):
  return proxy_settings.intercept.project_key

def __scenario_key(proxy_settings: ProxySettings):
  data_rule = __data_rule(proxy_settings)
  return data_rule.scenario_key


import pdb

from stoobly_agent.app.models.request_model import RequestModel
from stoobly_agent.app.models.schemas.request import Request
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.api.interfaces import RequestsIndexQueryParams
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey

from ...cli.helpers.context import ReplayContext
from .replay_request_service import replay_with_trace, ReplayRequestOptions
from .trace_context import TraceContext

PAGE_SIZE = '25'

def inject_replay():
  settings = Settings.instance()
  request_model = RequestModel(settings)
  endpoints_resource = EndpointsResource(settings.remote.api_url, settings.remote.api_key)

  return lambda scenario_key, options: replay(scenario_key, request_model, endpoints_resource, options)

def replay(
  scenario_key: str, request_model: RequestModel, endpoints_resource: EndpointsResource, options: ReplayRequestOptions
):
  scenario_key = ScenarioKey(scenario_key)
  trace_context = options.get('trace_context') or TraceContext(endpoints_resource)

  page = 0
  count = 0

  while True:
    requests_index, status = __get_requests(
      request_model, scenario_key, { 'page': page, 'size': PAGE_SIZE, 'sort_order': 'asc' }
    )

    if status != 200:
      # TODO: log error
      return

    if not requests_index:
      return

    requests = list(map(lambda request: Request(request), requests_index['list']))
    total = requests_index['total']
  
    if not requests or len(requests) == 0:
      return

    for request_partial in requests:
      count += 1

      request_id = request_partial.id
      request_response, status = __get_request(request_model, scenario_key, request_id)
      if not request_response:
        continue

      if status != 200:
        # TODO: log error
        return
      
      context = ReplayContext(Request(request_response)).with_sequence(count)
      replay_with_trace(context, trace_context, options)

    if count >= total:
      return

    page += 1

def __get_requests(
  request_model: RequestModel, scenario_key: ScenarioKey, query_params: RequestsIndexQueryParams = {}
):
  project_id = scenario_key.project_id
  scenario_id = scenario_key.id

  return request_model.index(**{
    'project_id': project_id,
    'scenario_id': scenario_id,
    'page': query_params.get('page'),
    'size': query_params.get('size'),
    'sort_order': query_params.get('sort_order'),
  })

def __get_request(request_model: RequestModel, scenario_key: ScenarioKey, request_id: int):
  project_id = scenario_key.project_id

  return request_model.show(
    request_id,
    **{
      'body': True,
      'headers': True,
      'project_id': project_id,
      'query_params': True,
      'response': True
    }
  )


from stoobly_agent.app.settings import RemoteSettings
from stoobly_agent.lib.orm.request import Request

from .local_db.body_adapter import LocalDBBodyAdapter

class BodyResourceFactory():

  def __init__(self, settings: RemoteSettings):
    self.__remote_settings = settings

  def local_db(self) -> LocalDBBodyAdapter:
    return LocalDBBodyAdapter(Request)  


# Wraps the .config.yml file in the workflow folder
import pdb

from .config import Config
from .constants import WORKFLOW_TEMPLATE_ENV

class WorkflowConfig(Config):

  def __init__(self, dir: str, **kwargs):
    super().__init__(dir)

    self.__template = None

    self.load()

    if 'template' in kwargs:
      self.__template = kwargs.get('template')

  @property
  def empty(self):
    return not self.template

  @property
  def template(self):
    return (self.__template or '').strip()

  @template.setter
  def template(self, v):
    self.__template = v

  def load(self, config = None):
    config = config or self.read()

    self.template = config.get(WORKFLOW_TEMPLATE_ENV)

  def to_dict(self):
    return {
      'template': self.template,
    }

  def write(self):
    config = {}

    if self.template:
      config[WORKFLOW_TEMPLATE_ENV] = self.template

    super().write(config)


type_map = {}
type_map[str(list)] = 'Array'
type_map[str(dict)] = 'Hash'
type_map[str(int)] = 'Integer'
type_map[str(float)] = 'Float'
type_map[str(str)] = 'String'
type_map[str(bool)] = 'Boolean'

reverse_type_map = {}
reverse_type_map['Array'] = list()
reverse_type_map['Hash'] = dict()
reverse_type_map['Integer'] = int()
reverse_type_map['Float'] = float()
reverse_type_map['String'] = str()
reverse_type_map['Boolean'] = bool() 

def convert(python_type: str):
  return type_map.get(python_type)

def convert_reverse(ruby_type: str):
  return reverse_type_map.get(ruby_type)




from typing import List, TypedDict

from .pagination_query_params import PaginationQueryParams

class ProjectsIndexQueryParams(PaginationQueryParams):
  organization_id: str

class ProjectDetails(TypedDict):
  name: str
  description: str

class ProjectCreateParams(ProjectDetails):
  organization_id: str
  project: ProjectDetails

class ProjectShowResponse(ProjectDetails):
  created_at: str
  key: str
  updated_at: str

class ProjectsIndexResponse(TypedDict):
  list: List[ProjectShowResponse]
  total: int



import base64
import json
import pdb

from typing import List, Tuple
from urllib.parse import parse_qs, urlencode, urlparse

from stoobly_agent.lib.api.interfaces import QueryParam, QueryParamShowResponse
from stoobly_agent.lib.orm.request import Request

from .local_db_adapter import LocalDBAdapter
from .request_adapter import LocalDBRequestAdapter

class LocalDBQueryParamAdapter(LocalDBAdapter):
  __request_orm = None

  def __init__(self, request_orm: Request.__class__ = Request):
    self.__request_orm = request_orm

  def create(self, request_id, **params: QueryParam):
    request = self.__request_orm.find(request_id)
    
    if not request:
      return self.__request_not_found()

    parsed_url = urlparse(request.url)

    _query_params = parse_qs(parsed_url.query)
    if not params['name'] in _query_params:
      _query_params[params['name']] = []

    if params['value'] not in _query_params[params['name']]:
      _query_params[params['name']].append(params['value'])
      parsed_url = parsed_url._replace(query=urlencode(_query_params, True))

    request = LocalDBRequestAdapter(self.__request_orm).update(request_id, url=parsed_url.geturl())

    return self.success({
      'name': params['name'],
      'value': params['value'],
    })

  def update(self, request_id, query_param_id, **params: QueryParam) -> Tuple[QueryParamShowResponse, int]:
    request = self.__request_orm.find(request_id)
    
    if not request:
      return self.__request_not_found()

    decoded_id = self.__decode_id(query_param_id)
    if not decoded_id:
      return self.bad_request('Invalid id')

    name = params['name']
    value = params['value']
    parsed_url = urlparse(request.url)

    _query_params = parse_qs(parsed_url.query)
    if not name in _query_params:
      _query_params[name] = []

    if isinstance(_query_params[name], list):
      try:
        index = _query_params[name].index(decoded_id['value'])
      except ValueError as e:
        return self.not_found()

      _query_params[name][index] = value
    else:
      _query_params[name] = value

    parsed_url = parsed_url._replace(query=urlencode(_query_params, True))

    request = LocalDBRequestAdapter(self.__request_orm).update(request_id, url=parsed_url.geturl())

    return self.success({
      'name': name,
      'value': value,
    })

  def index(self, request_id, **query_params) -> Tuple[List[QueryParamShowResponse], int]:
    request = self.__request_orm.find(request_id)

    if not request:
      return self.__request_not_found()

    request_dict = request.to_dict()
    parsed_query = parse_qs(request_dict['query'])

    query_params = []
    for key, params in parsed_query.items():
      for param in params:
        query_params.append({
          'name': key,
          'value': param,
        })

    return self.success(query_params)

  def destroy(self, request_id, id):
    request = self.__request_orm.find(request_id)
    
    if not request:
      return self.__request_not_found()

    decoded_id = self.__decode_id(id)
    if not decoded_id or not decoded_id.get('value'):
      return self.bad_request('Invalid id')

    parsed_url = urlparse(request.url)
    _query_params = parse_qs(parsed_url.query)

    name = decoded_id['name']
    value = decoded_id['value']

    try:
      index = _query_params[name].index(value)
    except ValueError as e:
      return self.not_found()

    del _query_params[name][index]

    parsed_url = parsed_url._replace(query=urlencode(_query_params, True))
    request = LocalDBRequestAdapter(self.__request_orm).update(request_id, url=parsed_url.geturl())

    return {
      'name': name,
      'value': value,
    }

  def __decode_id(self, id: str):
    id = base64.b64decode(id)

    try:
      id = json.loads(id)
    except Exception as e:
      return None

    if not isinstance(id, dict):
      return None

    if not id.get('name'):
      return None

    if not id.get('value'):
      return None

    return id

  def __request_not_found(self):
    return self.not_found('Request not found')


import json
import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.config.constants import mode
from stoobly_agent.app.settings.constants import request_component
from stoobly_agent.cli import config, mock, record
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

class TestMocking():
  class TestRewriting():

    @pytest.fixture(autouse=True, scope='class')
    def settings(self):
      return reset()

    class TestWhenHeaders():
      header_name = 'foo'
      header_value = 'bar'
      url = DETERMINISTIC_GET_REQUEST_URL

      @pytest.fixture(autouse=True, scope='class')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [self.url, '-H', f"{self.header_name}: {self.header_value}"])
        assert record_result.exit_code == 0
        return Request.last()

      def test_it_ignores_headers_default(self, runner: CliRunner):
        # Expect to fail since we have not set up rewrite rule for mock mode
        mock_result = runner.invoke(mock, [self.url])
        assert mock_result.exit_code == 0

      def test_it_fails_when_enabled(self, runner: CliRunner):
        # Enforce matching by headers only
        match_result = runner.invoke(config, [
          'match', 'set',
          '--method', 'GET', '--mode', mode.MOCK, '--pattern', '.*?', '--component', request_component.HEADER
        ])
        assert match_result.exit_code == 0
        
        # Expect to fail since we have not set up rewrite rule for mock mode
        mock_result = runner.invoke(mock, [self.url])
        assert mock_result.exit_code == 1

      def test_it_succeeds_after_rewrite(self, runner: CliRunner):
        rewrite_result = runner.invoke(config, [
            'rewrite', 'set', 
            '--method', 'GET', '--mode', mode.MOCK, '--name', self.header_name, '--value', self.header_value, '--pattern', '.*?', '--type', request_component.HEADER
          ]
        )
        assert rewrite_result.exit_code == 0

        mock_result = runner.invoke(mock, [self.url])
        assert mock_result.exit_code == 0

    class TestWhenQueryParams():
      query_param_name = 'foo'
      query_param_value = 'bar'
      url = DETERMINISTIC_GET_REQUEST_URL

      @pytest.fixture(autouse=True, scope='class')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [f"{self.url}?{self.query_param_name}={self.query_param_value}"])
        assert record_result.exit_code == 0
        return Request.last()

      def test_it_ignores_query_params_by_default(self, runner: CliRunner):
        mock_result = runner.invoke(mock, [self.url])
        assert mock_result.exit_code == 0

      def test_it_fails_when_enabled(self, runner: CliRunner):
        # Enforce matching by query params only
        match_result = runner.invoke(config, [
          'match', 'set',
          '--method', 'GET', '--mode', mode.MOCK, '--pattern', '.*?', '--component', request_component.QUERY_PARAM
        ])
        assert match_result.exit_code == 0

        # Expect to fail since we have not set up rewrite rule for mock mode
        mock_result = runner.invoke(mock, [self.url])
        assert mock_result.exit_code == 1

      def test_it_succeeds_after_rewrites(self, runner: CliRunner):
        rewrite_result = runner.invoke(config, [
            'rewrite', 'set', 
            '--method', 'GET', '--mode', mode.MOCK, '--name', self.query_param_name, '--value', self.query_param_value, '--pattern', '.*?', '--type', request_component.QUERY_PARAM
          ]
        )
        assert rewrite_result.exit_code == 0

        mock_result = runner.invoke(mock, [self.url])
        assert mock_result.exit_code == 0

    class TestWhenBodyParams():
      body_param_name = 'foo'
      body_param_value = 'bar'
      url = DETERMINISTIC_GET_REQUEST_URL

      @pytest.fixture(autouse=True, scope='class')
      def recorded_request(self, runner: CliRunner):
        content = {}
        content[self.body_param_name] = self.body_param_value
        body = json.dumps(content)

        url = DETERMINISTIC_GET_REQUEST_URL
        record_result = runner.invoke(record, [url, '-d', body, '-H', 'Content-Type: application/json'])
        assert record_result.exit_code == 0
        return Request.last()

      def test_it_ignores_body_params_by_default(self, runner: CliRunner):
        mock_result = runner.invoke(mock, [self.url])
        assert mock_result.exit_code == 0

      def test_it_fails_when_enabled(self, runner: CliRunner):
        # Enforce matching by bodyy params only
        match_result = runner.invoke(config, [
          'match', 'set',
          '--method', 'GET', '--mode', mode.MOCK, '--pattern', '.*?', '--component', request_component.BODY_PARAM
        ])
        assert match_result.exit_code == 0

        # Expect to fail since we have not set up rewrite rule for mock mode
        mock_result = runner.invoke(mock, [self.url])
        assert mock_result.exit_code == 1

      def test_it_succeeds_after_rewrites(self, runner: CliRunner):
        rewrite_result = runner.invoke(config, [
            'rewrite', 'set', 
            '--method', 'GET', '--mode', mode.MOCK, '--name', self.body_param_name, '--value', self.body_param_value, '--pattern', '.*?', '--type', request_component.BODY_PARAM
          ]
        )
        assert rewrite_result.exit_code == 0

        mock_result = runner.invoke(mock, [self.url])
        assert mock_result.exit_code == 0




from stoobly_agent.app.models.adapters import JoinedRequestAdapter
from stoobly_agent.app.models.adapters.python import PythonResponseAdapterFactory
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter

def verify_raw_request(raw_request: bytes):
  adapter = JoinedRequestAdapter(raw_request)
  response_string = adapter.build_response_string().get()
  python_response = RawHttpResponseAdapter(response_string).to_response()

  raw_response = PythonResponseAdapterFactory(python_response).raw_response()
  adapter.raw_response_string = raw_response

  joined_request = adapter.adapt()
  return joined_request.build()


endpoint_show_response = {
  "aliases": [
    {
      "id": 106,
      "name": ":endpointId",
      "created_at": "2022-05-31T07:03:40.975Z",
      "updated_at": "2022-05-31T07:03:40.975Z",
      "project_id": 37,
      "path_segment_names_count": 1,
      "header_names_count": 0,
      "query_param_names_count": 1,
      "body_param_names_count": 0,
      "response_param_names_count": 3
    },
    {
      "id": 107,
      "name": ":endpointRequestsCount",
      "created_at": "2022-05-31T07:03:40.975Z",
      "updated_at": "2022-05-31T07:03:40.975Z",
      "project_id": 37,
      "path_segment_names_count": 1,
      "header_names_count": 0,
      "query_param_names_count": 1,
      "body_param_names_count": 0,
      "response_param_names_count": 3
    }
  ],
  "query_param_names": [
    {
      "created_at": "2022-05-31T07:48:32.700Z",
      "id": 1,
      "inferred_type": "Integer",
      "is_deterministic": True,
      "is_required": False,
      "name": "limit",
      "query": "limit",
      "requests_count": 1,
      "query_param_name_id": None,
      "updated_at": "2022-05-31T07:48:32.700Z",
    },
    {
      "created_at": "2022-05-31T07:48:32.700Z",
      "id": 2,
      "inferred_type": "String",
      "is_deterministic": True,
      "is_required": False,
      "name": "sort",
      "query": "sort",
      "requests_count": 1,
      "query_param_name_id": None,
      "updated_at": "2022-05-31T07:48:32.700Z",
    },
    {
      "created_at": "2022-05-31T07:48:32.700Z",
      "id": 3,
      "inferred_type": "Array",
      "is_deterministic": True,
      "is_required": False,
      "name": "tags",
      "query": "tags",
      "requests_count": 1,
      "query_param_name_id": None,
      "updated_at": "2022-05-31T07:48:32.700Z",
    },
    {
      "created_at": "2022-05-31T07:48:32.700Z",
      "id": 4,
      "inferred_type": "String",
      "is_deterministic": True,
      "is_required": False,
      "name": "TagsElement",
      "query": "tags[*]",
      "requests_count": 1,
      "query_param_name_id": 3,
      "updated_at": "2022-05-31T07:48:32.700Z",
    },
  ],
  "body_param_names": [
    {
      "created_at": "2022-05-31T07:48:32.700Z",
      "id": 1,
      "inferred_type": "String",
      "is_deterministic": True,
      "is_required": True,
      "name": "name",
      "query": "name",
      "requests_count": 1,
      "body_param_name_id": None,
      "updated_at": "2022-05-31T07:48:32.700Z",
    },
  ],
  "response_param_names": [
    {
      "created_at": "2022-05-31T07:48:32.700Z",
      "id": 431,
      "inferred_type": "Integer",
      "is_deterministic": True,
      "is_required": True,
      "name": "requests_count",
      "query": "[*].requests_count",
      "requests_count": 2,
      "response_param_name_id": 426,
      "updated_at": "2022-05-31T07:48:32.700Z",
      "alias_id": 107,
    },
    {
      "created_at": "2022-05-31T07:48:32.722Z",
      "id": 433,
      "inferred_type": "Integer",
      "is_deterministic": True,
      "is_required": True,
      "name": "category",
      "query": "[*].category",
      "requests_count": 2,
      "response_param_name_id": 426,
      "updated_at": "2022-05-31T07:48:32.722Z"
    },
    {
      "created_at": "2022-05-31T07:48:32.780Z",
      "id": 440,
      "inferred_type": "String",
      "is_deterministic": False,
      "is_required": True,
      "name": "created_at",
      "query": "[*].created_at",
      "requests_count": 2,
      "response_param_name_id": 426,
      "updated_at": "2022-06-15T01:07:41.094Z"
    },
    {
      "created_at": "2022-05-31T07:48:32.864Z",
      "id": 444,
      "inferred_type": "String",
      "is_deterministic": True,
      "is_required": True,
      "name": "url",
      "query": "[*].url",
      "requests_count": 2,
      "response_param_name_id": 426,
      "updated_at": "2022-05-31T07:48:32.864Z"
    },
    {
      "created_at": "2022-05-31T07:48:32.742Z",
      "id": 435,
      "inferred_type": "String",
      "is_deterministic": True,
      "is_required": True,
      "name": "path",
      "query": "[*].path",
      "requests_count": 2,
      "response_param_name_id": 426,
      "updated_at": "2022-05-31T07:48:32.742Z"
    },
    {
      "created_at": "2022-05-31T07:48:32.760Z",
      "id": 437,
      "inferred_type": "String",
      "is_deterministic": True,
      "is_required": True,
      "name": "method",
      "query": "[*].method",
      "requests_count": 2,
      "response_param_name_id": 426,
      "updated_at": "2022-05-31T07:48:32.760Z"
    },
    {
      "created_at": "2022-05-31T07:48:32.812Z",
      "id": 442,
      "inferred_type": "String",
      "is_deterministic": False,
      "is_required": True,
      "name": "updated_at",
      "query": "[*].updated_at",
      "requests_count": 2,
      "response_param_name_id": 426,
      "updated_at": "2022-06-15T01:07:43.561Z"
    },
    {
      "created_at": "2022-05-31T07:48:32.834Z",
      "id": 443,
      "inferred_type": "Array",
      "is_deterministic": True,
      "is_required": True,
      "name": "components",
      "query": "[*].components",
      "requests_count": 2,
      "response_param_name_id": 426,
      "updated_at": "2022-05-31T07:48:32.834Z"
    },
    {
      "created_at": "2022-05-31T07:48:32.679Z",
      "id": 428,
      "inferred_type": "Integer",
      "is_deterministic": False,
      "is_required": True,
      "name": "id",
      "query": "[*].id",
      "requests_count": 2,
      "response_param_name_id": 426,
      "updated_at": "2022-06-15T01:08:06.946Z",
      "alias_id": 106, 
    },
    {
      "created_at": "2022-05-31T07:48:32.635Z",
      "id": 426,
      "inferred_type": "Hash",
      "is_deterministic": True,
      "is_required": True,
      "name": "Element",
      "query": "[*]",
      "requests_count": 2,
      "response_param_name_id": None,
      "updated_at": "2022-06-06T08:21:50.835Z"
    }
  ]
}



import json
import pdb

from stoobly_agent.app.proxy.record import REQUEST_DELIMITTER

from .upload_test_service import UploadTestData

JSON_DATA = b'json'
TEXT_DATA = b'text'

class DownloadTestData(UploadTestData):
  received_response: bytes

class TestResultsBuilder():

  def __init__(self, **test_data: DownloadTestData):
    self.__expected_latency = test_data.get('expected_latency')
    self.__expected_response = test_data.get('expected_response')
    self.__expected_status_code = test_data.get('expected_status_code')
    self.__log = test_data.get('log')
    self.__passed = test_data.get('passed')
    self.__request_id = test_data.get('request_id')
    self.__received_response = test_data.get('received_response')
    self.__skipped = test_data.get('skipped')
    self.__status = test_data.get('status')
    self.__strategy = test_data.get('strategy')

  @property
  def expected_latency(self):
    return self.__expected_latency

  @property
  def expected_response(self):
    return self.__expected_response

  @property
  def expected_status_code(self):
    return self.__expected_status_code

  @property
  def log(self):
    return self.__log

  @property
  def passed(self):
    return self.__passed

  @property
  def request_id(self):
    return self.__request_id

  @property
  def received_response(self):
    return self.__received_response

  @property
  def skipped(self):
    return self.__skipped

  @property
  def status(self):
    return self.__status

  @property
  def strategy(self):
    return self.__strategy

  def with_expected_latency(self, latency: int):
    self.__expected_latency = latency
    return self

  def with_expected_response(self, response: bytes):
    self.__expected_response = response
    return self

  def with_expected_status_code(self, status_code: int):
    self.__expected_status_code = status_code
    return self

  def with_log(self, log):
    self.__log = log
    return self

  def with_passed(self, passed: bool):
    self.__passed = passed
    return self

  def with_received(self, response: bytes):
    self.__received_response = response
    return self

  def with_request_id(self, request_id):
    self.__request_id = request_id
    return self

  def with_status(self, status):
    self.__status = status
    return self

  def with_strategy(self, strategy):
    self.__strategy = strategy
    return self

  def serialize(self) -> bytes:
    metadata = {
      'expected_latency': self.__expected_latency, 
      'expected_status_code': self.__expected_status_code, 
      'log': self.__log,
      'passed': self.__passed,
      'request_id': self.__request_id,
      'skipped': self.__skipped,
      'status': self.__status,
      'strategy': self.__strategy, 
    }

    received, received_format = self.__serialize_response(self.__received_response)
    expected, expected_format = self.__serialize_response(self.__expected_response)
    return REQUEST_DELIMITTER.join([
      received_format,
      received,
      expected_format,
      expected,
      json.dumps(metadata).encode()
    ])

  def deserialize(self, data: bytes):
    toks = data.split(REQUEST_DELIMITTER)

    toks_length = len(toks)

    self.__received_response_header = toks[toks_length - 5]
    self.__received_response = toks[toks_length - 4]
    if self.__received_response_header == JSON_DATA:
      self.__received_response = json.loads(self.__received_response.decode())

    self.__expected_response_header = toks[toks_length - 3]
    self.__expected_response = toks[toks_length - 2]
    if self.__expected_response_header == JSON_DATA:
      self.__expected_response = json.loads(self.__expected_response.decode())

    test_data = {} 
    try: 
      test_data = json.loads(toks[toks_length - 1].decode())
    except:
      pass
    
    self.__expected_latency = test_data.get('expected_latency')
    self.__expected_status_code = test_data.get('expected_status_code')
    self.__log = test_data.get('log')
    self.__passed = test_data.get('passed')
    self.__request_id = test_data.get('request_id')
    self.__skipped = test_data.get('skipped')
    self.__status = test_data.get('status')
    self.__strategy = test_data.get('strategy')

  def __serialize_response(self, data):
    if data == None:
      return b'', TEXT_DATA

    if isinstance(data, bytes):
      return data, TEXT_DATA 
      
    return json.dumps(data).encode(), JSON_DATA


from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow
from typing import List

from stoobly_agent.app.settings.constants.request_component import REQUEST_COMPONENTS, RESPONSE_COMPONENTS
from stoobly_agent.app.settings.rewrite_rule import RewriteRule

from ..mitmproxy.request_facade import MitmproxyRequestFacade
from ..mitmproxy.response_facade import MitmproxyResponseFacade

def select_request_rewrite_rules(rewrite_rules: List[RewriteRule]):
  rules = []

  for rewrite_rule in rewrite_rules:
    parameter_rules = list(filter(
      lambda parameter: parameter.type in REQUEST_COMPONENTS and parameter.name, 
      rewrite_rule.parameter_rules or []
    ))

    if len(parameter_rules) > 0 or len(rewrite_rule.url_rules) > 0:
      rewrite_rule = RewriteRule(rewrite_rule.to_dict())
      rewrite_rule.url_rules = rewrite_rule.url_rules
      rewrite_rule.parameter_rules = parameter_rules
      rules.append(rewrite_rule)

  return rules

def select_response_rewrite_rules(rewrite_rules: List[RewriteRule]):
  rules = []

  for rewrite_rule in rewrite_rules:
    parameter_rules = list(filter(
      lambda parameter: parameter.type in RESPONSE_COMPONENTS and parameter.name, 
      rewrite_rule.parameter_rules or []
    ))

    if len(parameter_rules) > 0:
      rewrite_rule = RewriteRule(rewrite_rule.to_dict())
      rewrite_rule.url_rules = [] 
      rewrite_rule.parameter_rules = parameter_rules
      rules.append(rewrite_rule)

  return rules

def rewrite_request_response(flow: MitmproxyHTTPFlow, rewrite_rules: List[RewriteRule]):
  request = rewrite_request(flow, rewrite_rules)
  response = rewrite_response(flow, rewrite_rules, request)
  return request, response

def rewrite_request(flow: MitmproxyHTTPFlow, rewrite_rules: List[RewriteRule]):
  request = None

  # Adapt flow.request
  request = MitmproxyRequestFacade(flow.request)

  _rewrite_rules = select_request_rewrite_rules(rewrite_rules) 
  if len(_rewrite_rules):
    request.with_parameter_rules(_rewrite_rules).with_url_rules(_rewrite_rules).rewrite()

  return request

def rewrite_response(flow: MitmproxyHTTPFlow, rewrite_rules: List[RewriteRule], request = None):
  # Adapt flow.request
  request = request or MitmproxyRequestFacade(flow.request)

  # Adapt flow.response
  response = MitmproxyResponseFacade(flow.response)

  _rewrite_rules = select_response_rewrite_rules(rewrite_rules)
  if len(_rewrite_rules):
    response.with_parameter_rules(_rewrite_rules, request).rewrite()

  return response


import pdb

from docker import errors as docker_errors

from stoobly_agent.app.cli.scaffold.constants import WORKFLOW_TEST_TYPE
from stoobly_agent.app.cli.scaffold.managed_services_docker_compose import (
  ManagedServicesDockerCompose,
)
from stoobly_agent.app.cli.scaffold.templates.constants import (
  CORE_ENTRYPOINT_SERVICE_NAME,
  CORE_GATEWAY_SERVICE_NAME,
  CORE_MOCK_UI_SERVICE_NAME,
  CORE_SERVICES,
)
from stoobly_agent.app.cli.scaffold.validate_command import ValidateCommand
from stoobly_agent.app.cli.scaffold.validate_exceptions import ScaffoldValidateException
from stoobly_agent.app.cli.scaffold.workflow_command import WorkflowCommand
from stoobly_agent.lib.logger import bcolors

from .app import App


class WorkflowValidateCommand(WorkflowCommand, ValidateCommand):
  def __init__(self, app: App, **kwargs):
    WorkflowCommand.__init__(self, app, **kwargs)
    ValidateCommand.__init__(self)
    self.managed_services_docker_compose = ManagedServicesDockerCompose(target_workflow_name=self.workflow_name)

  # Gateway core service runs in all workflows
  def validate_gateway_service(self):
    print(f"Validating core service: {CORE_GATEWAY_SERVICE_NAME}")
    gateway_container_name = self.managed_services_docker_compose.gateway_container_name

    container_missing_message = f"{bcolors.FAIL}Container '{gateway_container_name}' not found for service '{CORE_GATEWAY_SERVICE_NAME}'{bcolors.ENDC}"
    suggestion_message = f"{bcolors.BOLD}Workflow might not be running yet. Try running 'scaffold workflow up <WORKFLOW_NAME>'{bcolors.ENDC}"
    error_message = f"{container_missing_message}\n\n{suggestion_message}"

    try:
      gateway_container = self.docker_client.containers.get(gateway_container_name)
      if not gateway_container or (gateway_container.status != 'running'):
        raise ScaffoldValidateException(error_message)
    except docker_errors.NotFound:
      raise ScaffoldValidateException(error_message)

  def validate_mock_ui_service(self):
    mock_ui_container_name = self.managed_services_docker_compose.mock_ui_container_name

    # The stoobly-ui service does not run in test workflows
    if self.workflow_name == WORKFLOW_TEST_TYPE:
      try:
        mock_ui_container = self.docker_client.containers.get(mock_ui_container_name)

        if mock_ui_container:
          ui_found_message = f"{bcolors.FAIL}Stoobly UI container is running when it shouldn't: {mock_ui_container_name}{bcolors.ENDC}"
          suggestion_message = f"{bcolors.BOLD}Run 'docker ps | grep stoobly_ui' to check if the Stoobly UI is running during the test workflow. Did you stop the record or mock workflow yet?{bcolors.ENDC}"
          error_message = f"{ui_found_message}\n\n{suggestion_message}"
          raise ScaffoldValidateException(error_message)
      except docker_errors.NotFound:
        print(f"Skipping validating core service: {CORE_MOCK_UI_SERVICE_NAME}")
        return

    print(f"Validating core service: {CORE_MOCK_UI_SERVICE_NAME}")
    container_missing_message = f"{bcolors.FAIL}Container '{mock_ui_container_name}' not found for service '{CORE_MOCK_UI_SERVICE_NAME}'{bcolors.ENDC}"
    suggestion_message = f"{bcolors.BOLD}Stoobly UI is not running. Check if the container is up with 'docker ps -a | grep {mock_ui_container_name}'{bcolors.ENDC}"
    mock_ui_missing_error_message = f"{container_missing_message}\n\n{suggestion_message}"

    try:
      mock_ui_container = self.docker_client.containers.get(mock_ui_container_name)
      if not mock_ui_container or (mock_ui_container.status != 'running'):
        raise ScaffoldValidateException(mock_ui_missing_error_message)
    except docker_errors.NotFound:
      raise ScaffoldValidateException(mock_ui_missing_error_message)

  def validate_entrypoint_service(self):
    print(f"Validating core service: {CORE_ENTRYPOINT_SERVICE_NAME}")

    core_entrypoint_init_container_name = None
    try:
      core_entrypoint_init_container_name = self.managed_services_docker_compose.entrypoint_init_container_name
      entrypoint_init_container = self.docker_client.containers.get(core_entrypoint_init_container_name)
    except docker_errors.NotFound:
      error_message = self._ValidateCommand__generate_container_not_found_error(core_entrypoint_init_container_name)
      raise ScaffoldValidateException(error_message)
    
    core_entrypoint_configure_container_name = None
    try:
      core_entrypoint_configure_container_name = self.managed_services_docker_compose.entrypoint_configure_container_name
      entrypoint_configure_container = self.docker_client.containers.get(core_entrypoint_configure_container_name)
    except docker_errors.NotFound:
      error_message = self._ValidateCommand__generate_container_not_found_error(core_entrypoint_configure_container_name)
      raise ScaffoldValidateException(error_message)

    # NOTE: we should check the correct workflow mode is enabled one day
    # That's not currently queryable

    print(f"{bcolors.OKGREEN}✔ Done validating core services for workflow: {self.workflow_name}, success!\n{bcolors.ENDC}")

  def validate_core_services(self):
    self.validate_gateway_service()
    self.validate_mock_ui_service()
    self.validate_init_containers(self.managed_services_docker_compose.init_container_name, self.managed_services_docker_compose.configure_container_name)
    self.validate_entrypoint_service()

  def validate(self) -> bool:
    print(f"Validating workflow: {self.workflow_name}\n")
    print(f"Validating core services: {CORE_SERVICES}")

    self.validate_core_services()

    return True





from stoobly_orator.migrations import Migration


class CreateTraces(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.create('traces') as table:
            table.increments('id')
            table.timestamps()

    def down(self):
        """
        Revert the migrations.
        """
        self.schema.drop('traces')



import os

CORE_BUILD_SERVICE_NAME = 'build'
CORE_ENTRYPOINT_SERVICE_NAME = 'entrypoint'
CORE_GATEWAY_SERVICE_NAME = 'gateway'
CORE_MOCK_UI_SERVICE_NAME = 'stoobly-ui'
CORE_SERVICES = [
  CORE_BUILD_SERVICE_NAME, CORE_ENTRYPOINT_SERVICE_NAME, CORE_MOCK_UI_SERVICE_NAME, CORE_GATEWAY_SERVICE_NAME 
]

CORE_MOCK_WORKFLOW = 'mock'
CORE_RECORD_WORKFLOW = 'record'

CUSTOM_BUILD = os.path.join('bin', 'build')
CUSTOM_CONFIGURE = os.path.join('bin', 'configure')
CUSTOM_INIT = os.path.join('bin', 'init')
CUSTOM_FIXTURES = 'fixtures.yml'
CUSTOM_LIFECYCLE_HOOKS = os.path.join('lifecycle_hooks.py')
CUSTOM_PUBLIC_GITIGNORE = os.path.join('public', '.gitignore')
MAINTAINED_CONFIGURE = os.path.join('bin', '.configure')
MAINTAINED_INIT = os.path.join('bin', '.init')

MOCK_WORKFLOW_MAINTAINED_FILES = [
  MAINTAINED_CONFIGURE,
  MAINTAINED_INIT,
]

MOCK_WORKFLOW_CUSTOM_FILES = [
  CUSTOM_BUILD,
  CUSTOM_CONFIGURE,
  CUSTOM_FIXTURES,
  CUSTOM_INIT,
  CUSTOM_LIFECYCLE_HOOKS,
  CUSTOM_PUBLIC_GITIGNORE
]

RECORD_WORKFLOW_MAINTAINED_FILES = [
  MAINTAINED_CONFIGURE,
  MAINTAINED_INIT
]

RECORD_WORKFLOW_CUSTOM_FILES = [
  CUSTOM_BUILD,
  CUSTOM_CONFIGURE,
  CUSTOM_INIT,
  CUSTOM_LIFECYCLE_HOOKS,
]

TEST_WORKFLOW_MAINTAINED_FILES = [
  MAINTAINED_CONFIGURE,
  MAINTAINED_INIT,
]

TEST_WORKFLOW_CUSTOM_FILES = [
  CUSTOM_BUILD,
  CUSTOM_CONFIGURE,
  CUSTOM_FIXTURES,
  CUSTOM_INIT,
  CUSTOM_LIFECYCLE_HOOKS,
  CUSTOM_PUBLIC_GITIGNORE
]

SERVICE_HOSTNAME_BUILD_ARG = 'SERVICE_HOSTNAME'



import pdb
import requests

from .mitmproxy_adapter import MitmproxyResponseAdapter
from .raw_adapter import RawResponseAdapter

class PythonResponseAdapterFactory():

  def __init__(self, response: requests.Response):
    self.__response = response

  def mitmproxy_response(self):
    return MitmproxyResponseAdapter(self.__response).adapt()

  def raw_response(self):
    return RawResponseAdapter(self.__response).adapt()


from stoobly_agent.app.proxy.context import InterceptContext
from stoobly_agent.config.constants.custom_headers import MOCK_REQUEST_ID

def handle_before_response(context: InterceptContext):
  print(context.flow.response.headers[MOCK_REQUEST_ID], end='')


import pdb

from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.scenario import Scenario

from .log import Log
from .log_event import Action, DELETE_ACTION, PUT_ACTION
from .request_snapshot import RequestSnapshot, RequestSnapshotOptions
from .scenario_snapshot import ScenarioSnapshot

def snapshot_request(request: Request, **options: RequestSnapshotOptions):
  action: Action = options.get('action')
  if not __validate_action(action):
    return

  snapshot = RequestSnapshot(request.uuid)

  snapshot.backup()

  if action == PUT_ACTION:
    snapshot.write(request, **options)
  elif action == DELETE_ACTION:
    snapshot.remove()

  log = Log()

  try:
    if action == PUT_ACTION:
      log.put(request)
    elif action == DELETE_ACTION:
      log.delete(request)
  except Exception as e:
    snapshot.rollback() 

  return snapshot.path

def snapshot_scenario(scenario: Scenario, **options):
  action: Action = options.get('action')
  if not __validate_action(action):
    return

  snapshot = ScenarioSnapshot(scenario.uuid)

  snapshot.backup_metadata()

  if action == PUT_ACTION:
    snapshot.write_metadata(scenario)
  elif action == DELETE_ACTION:
    snapshot.remove_metadata()

  snapshot.backup_requests()

  if action == PUT_ACTION:
    snapshot.remove_requests()
    snapshot.write_requests(scenario, **options)
  elif action == DELETE_ACTION:
    snapshot.remove_requests()

  log = Log()

  if action == PUT_ACTION:
    log.put(scenario)
  elif action == DELETE_ACTION:
    log.delete(scenario)

  return snapshot.metadata_path

def __validate_action(action: Action):
  return action == PUT_ACTION or action == DELETE_ACTION 


import pdb
import pytest

from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.app.proxy.test.matchers.context import MatchContext
from stoobly_agent.app.proxy.test.matchers.fuzzy import list_matches
from stoobly_agent.test.mock_data.endpoint_show_response import endpoint_show_response

@pytest.fixture
def endpoints_list_response_param_names_facade():
  response_param_names = endpoint_show_response['response_param_names']
  aliases = endpoint_show_response['aliases']
  return RequestComponentNamesFacade(response_param_names).with_aliases(aliases)
  
@pytest.fixture
def empty_response_param_names_facade():
  response_param_names = []
  return RequestComponentNamesFacade(response_param_names)

class TestMatchesFuzzySanity():
  def test_matches_empty_response(self, endpoints_list_response_param_names_facade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      [], []
    )
    assert matches, log

  def test_matches_list_of_numbers(self, empty_response_param_names_facade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': empty_response_param_names_facade
      }),
      [1], [2]
    )
    assert matches, log

  def test_not_matches_mixed(self, endpoints_list_response_param_names_facade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      [{}], [{}, 1]
    )
    
    assert log == "Key '[1]' type did not match: got <class 'int'>, expected valid types <class 'dict'>", log 

  def test_not_matches(self, endpoints_list_response_param_names_facade):
    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      [], {}
    )
    assert log == "Key '' type did not match: got <class 'dict'>, expected <class 'list'>", log 

class TestFuzzyMatchesListOfDicts():

  def test_matches_response_one_object(self, endpoints_list_response_param_names_facade):
    actual = [
      {
        "id": 1000,
        "requests_count": 8,
        "category": 2,
        "path": "/abc",
        "method": "POST",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": [],
        "url": "https://alpha.api.stoobly.com/endpoints/930"
      },
    ]

    expected = [
      {
        "id": 930,
        "requests_count": 6,
        "category": 1,
        "path": "/abc",
        "method": "GET",
        "created_at": "2022-05-26T23:46:49.968Z",
        "updated_at": "2022-05-26T23:46:49.968Z",
        "components": [],
        "url": "https://alpha.api.stoobly.com/endpoints/930"
      },
    ] 

    matches, log = list_matches(
      MatchContext({
        'path_key': '', 'query': '', 'request_component_names_facade': endpoints_list_response_param_names_facade
      }),
      expected, actual
    )
    assert matches, log


import os
import pdb
from typing import List

from .app import App


class Workflow():

  def __init__(self, workflow_name: str, app: App):
    self.__app = app
    self.__workflow_name = workflow_name

  @property
  def app(self):
    return self.__app

  @property
  def workflow_name(self):
    return self.__workflow_name

  @property
  def service(self):
    return self._service

  @service.setter
  def service(self, v):
    self._service = v

  @property
  def services(self):
    return list(map(lambda path: os.path.basename(path), self.service_paths))

  @property
  def service_paths(self):
    all_service_paths = self.app.service_paths
    return list(filter(lambda path: os.path.exists(os.path.join(path, self.workflow_name)), all_service_paths))

  # TODO: merge into 1 services property

  # Returns services that run in this specific workflow
  @property
  def services_ran(self) -> List[str]:
    services_dir = os.path.join(self.app.data_dir_path, self.app.scaffold_namespace)

    services = []
    for filename in os.listdir(services_dir):
      path = os.path.join(services_dir, filename)
      if not os.path.isdir(path):
        continue

      for sub_path in os.scandir(path):
        if os.path.isdir(sub_path):
          if sub_path.name == self.workflow_name:
            services.append(filename)

    return services

  def service_paths_from_services(self, services: List[str]):
    scaffold_namespace_path = self.app.scaffold_namespace_path
    return list(map(lambda service: os.path.join(scaffold_namespace_path, service), services))



import os

from stoobly_agent.config.constants import env_vars

from .data_settings import DataSettings
from .firewall_settings import FirewallSettings
from .intercept_settings import InterceptSettings
from .match_settings import MatchSettings
from .rewrite_settings import RewriteSettings
from .types.proxy_settings import ProxySettings as IProxySettings

class ProxySettings:

  def __init__(self, proxy_settings: IProxySettings):
    self.__proxy_settings = proxy_settings or {}

    self.__data = DataSettings(self.__proxy_settings.get('data'))
    self.__firewall = FirewallSettings(self.__proxy_settings.get('firewall'))
    self.__intercept = InterceptSettings(self.__proxy_settings.get('intercept'))
    self.__match = MatchSettings(self.__proxy_settings.get('match'))
    self.__rewrite = RewriteSettings(self.__proxy_settings.get('rewrite')) 
    self.__url = self.__proxy_settings.get('url')

  @property
  def data(self) -> DataSettings:
    return self.__data

  @property
  def rewrite(self) -> RewriteSettings:
    return self.__rewrite

  @property
  def firewall(self) -> FirewallSettings:
    return self.__firewall

  @property
  def intercept(self) -> InterceptSettings:
    return self.__intercept

  @property
  def match(self) -> MatchSettings:
    return self.__match

  @property
  def url(self) -> str:
    if os.environ.get(env_vars.AGENT_PROXY_URL):
        return os.environ[env_vars.AGENT_PROXY_URL]

    return self.__url or ''

  @url.setter
  def url(self, v):
    self.__url = v

  def to_dict(self) -> IProxySettings:
    _dict = {
      'data': self.__data.to_dict(),
      'firewall': self.__firewall.to_dict(),
      'intercept': self.__intercept.to_dict(),
      'match': self.__match.to_dict(),
      'rewrite': self.__rewrite.to_dict(),
    }

    if self.__url:
      _dict['url'] = self.__url

    return _dict


from typing import TypedDict

OPENAPI_FORMAT = 'openapi'

ENDPOINT_COMPONENT_NAMES = [
  "header_name",
  "body_param_name",
  "query_param_name",
  "response_header_name",
  "response_param_name"
]

class EndpointCreateParams(TypedDict):
  host: str
  method: str
  path_segments: str
  path: str
  port: str
  project_id: str

class HeaderNameCreateParams(TypedDict):
  name: str
  is_required: bool
  is_deterministic: bool
  project_id: str
  endpoint_id: int
	
class ParamNameCreateParams(TypedDict):
  name: str
  project_id: str
  endpoint_id: int
  inferred_type: str
  query: str


from pathlib import Path
import pdb
import shutil

from click.testing import CliRunner
import pytest

from stoobly_agent.app.cli.scaffold.app import App
from stoobly_agent.app.cli.scaffold.constants import (
  DOCKER_NAMESPACE,
  WORKFLOW_RECORD_TYPE,
  WORKFLOW_TEST_TYPE,
)
from stoobly_agent.app.cli.scaffold.managed_services_docker_compose import (
  ManagedServicesDockerCompose,
)
from stoobly_agent.app.cli.scaffold.service_docker_compose import ServiceDockerCompose
from stoobly_agent.app.cli.scaffold.service_workflow_validate_command import (
  ServiceWorkflowValidateCommand,
)
from stoobly_agent.app.cli.scaffold.workflow_validate_command import (
  WorkflowValidateCommand,
)
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.test.app.cli.scaffold.cli_invoker import ScaffoldCliInvoker
from stoobly_agent.test.test_helper import reset


@pytest.mark.e2e
class TestScaffoldE2e():

  @pytest.fixture(scope='module', autouse=True)
  def settings(self):
    return reset()

  @pytest.fixture(scope='module')
  def runner(self):
    yield CliRunner()

  @pytest.fixture(scope='class')
  def app_name(self):
    yield "0.0.1"

  @pytest.fixture(scope='class', autouse=True)
  def temp_dir(self):
    data_dir_path = DataDir.instance().path
    tmp_path = data_dir_path[:data_dir_path.rfind('/')]

    yield tmp_path

  @pytest.fixture(scope='class', autouse=True)
  def app_dir_path(self, temp_dir, app_name):
    yield temp_dir

  @pytest.fixture(scope='class')
  def mock_data_directory_path(self):
    yield Path(__file__).parent.parent.parent.parent / 'mock_data'
  
  @pytest.fixture(scope='class')
  def local_service_mock_docker_compose_path(self, mock_data_directory_path):
    path = mock_data_directory_path / "scaffold" / "docker-compose-local-service.yml"
    yield path

  @pytest.fixture(scope='class')
  def hostname(self):
    yield "http.badssl.com"

  @pytest.fixture(scope='class')
  def https_service_hostname(self):
    yield "example.com"
  
  @pytest.fixture(scope='class')
  def external_service_name(self):
    yield "external-service"

  @pytest.fixture(scope='class')
  def external_https_service_name(self):
    yield "external-https-service"

  @pytest.fixture(scope='class')
  def local_hostname(self):
    yield "my-httpbin.com"

  @pytest.fixture(scope='class')
  def local_service_name(self):
    yield "my-httpbin"

  
  class TestRecordWorkflow():
    @pytest.fixture(scope='class', autouse=True)
    def target_workflow_name(self):
      yield WORKFLOW_RECORD_TYPE

    @pytest.fixture(scope='class')
    def managed_services_docker_compose(self, target_workflow_name):
      yield ManagedServicesDockerCompose(target_workflow_name=target_workflow_name)

    @pytest.fixture(scope='class')
    def external_service_docker_compose(self, app_dir_path, target_workflow_name, external_service_name, hostname):
      yield ServiceDockerCompose(app_dir_path=app_dir_path, target_workflow_name=target_workflow_name, service_name=external_service_name, hostname=hostname)
    
    @pytest.fixture(scope='class')
    def external_https_service_docker_compose(self, app_dir_path, target_workflow_name, external_https_service_name, https_service_hostname):
      yield ServiceDockerCompose(app_dir_path=app_dir_path, target_workflow_name=target_workflow_name, service_name=external_https_service_name, hostname=https_service_hostname)

    @pytest.fixture(scope='class')
    def local_service_docker_compose(self, app_dir_path, target_workflow_name, local_service_name, local_hostname):
      yield ServiceDockerCompose(app_dir_path=app_dir_path, target_workflow_name=target_workflow_name, service_name=local_service_name, hostname=local_hostname)

    @pytest.fixture(scope="class", autouse=True)
    def create_scaffold_setup(self, runner, app_dir_path, app_name, target_workflow_name, external_service_docker_compose, external_https_service_docker_compose, local_service_docker_compose, local_service_mock_docker_compose_path):
      ScaffoldCliInvoker.cli_app_create(runner, app_dir_path, app_name)

      # Create external user defined service
      ScaffoldCliInvoker.cli_service_create(runner, app_dir_path, external_service_docker_compose.hostname, external_service_docker_compose.service_name, False)
      ScaffoldCliInvoker.cli_service_create(runner, app_dir_path, external_https_service_docker_compose.hostname, external_https_service_docker_compose.service_name, True)

      # Create local user defined service
      ScaffoldCliInvoker.cli_service_create(runner, app_dir_path, local_service_docker_compose.hostname, local_service_docker_compose.service_name, False)

      # Validate docker-compose path exists
      destination_path = Path(local_service_docker_compose.docker_compose_path)
      assert destination_path.is_file()
      # Add user defined Docker Compose file for the custom container service
      shutil.copyfile(local_service_mock_docker_compose_path, destination_path)

      # Record workflow doesn't have a public folder

      # Generate certs
      ScaffoldCliInvoker.cli_app_mkcert(runner, app_dir_path)

      ScaffoldCliInvoker.cli_workflow_up(runner, app_dir_path, target_workflow_name)

    @pytest.fixture(scope="class", autouse=True)
    def cleanup_after_all(self, runner, app_dir_path, target_workflow_name):
      yield
      ScaffoldCliInvoker.cli_workflow_down(runner, app_dir_path, target_workflow_name)
      shutil.rmtree(app_dir_path)

    def test_core_services(self, app_dir_path, target_workflow_name):
      app = App(app_dir_path, DOCKER_NAMESPACE)
      config = {
        'workflow_name': target_workflow_name,
        'service_name': 'build'
      }

      command = WorkflowValidateCommand(app, **config)
      command.validate()

    def test_external_service(self, external_service_docker_compose: ServiceDockerCompose, app_dir_path, target_workflow_name):
      app = App(app_dir_path, DOCKER_NAMESPACE)
      config = {
        'workflow_name': target_workflow_name,
        'service_name': external_service_docker_compose.service_name 
      }

      command = ServiceWorkflowValidateCommand(app, **config)
      command.validate()

    def test_local_service(self, app_dir_path, target_workflow_name, local_service_docker_compose: ServiceDockerCompose):
      app = App(app_dir_path, DOCKER_NAMESPACE)
      config = {
        'workflow_name': target_workflow_name,
        'service_name': local_service_docker_compose.service_name 
      }

      command = ServiceWorkflowValidateCommand(app, **config)
      command.validate()


  class TestTestWorkflow():
    @pytest.fixture(scope='class')
    def assets_service_mock_docker_compose_path(self, mock_data_directory_path):
      path = mock_data_directory_path / "scaffold" / "docker-compose-assets-service.yml"
      yield path

    @pytest.fixture(scope='class')
    def assets_service_assets_path(self, mock_data_directory_path):
      path = mock_data_directory_path / "scaffold" / "index.html"
      yield path
    
    @pytest.fixture(scope='class', autouse=True)
    def target_workflow_name(self):
      yield WORKFLOW_TEST_TYPE

    @pytest.fixture(scope='class')
    def assets_service_name(self):
      yield "assets"
    
    @pytest.fixture(scope='class')
    def assets_hostname(self):
      yield "assets"

    @pytest.fixture(scope='class')
    def managed_services_docker_compose(self, target_workflow_name):
      yield ManagedServicesDockerCompose(target_workflow_name=target_workflow_name)
    
    @pytest.fixture(scope='class')
    def external_service_docker_compose(self, app_dir_path, target_workflow_name, external_service_name, hostname):
      yield ServiceDockerCompose(app_dir_path=app_dir_path, target_workflow_name=target_workflow_name, service_name=external_service_name, hostname=hostname)

    @pytest.fixture(scope='class')
    def local_service_docker_compose(self, app_dir_path, target_workflow_name, local_service_name, local_hostname):
      yield ServiceDockerCompose(app_dir_path=app_dir_path, target_workflow_name=target_workflow_name, service_name=local_service_name, hostname=local_hostname)
    
    @pytest.fixture(scope='class')
    def assets_service_docker_compose(self, app_dir_path, target_workflow_name, assets_service_name, assets_hostname):
      yield ServiceDockerCompose(app_dir_path=app_dir_path, target_workflow_name=target_workflow_name, service_name=assets_service_name, hostname=assets_hostname)

    @pytest.fixture(scope='class', autouse=True)
    def setup_docker_composes(self, managed_services_docker_compose, external_service_docker_compose, local_service_docker_compose, assets_service_docker_compose):
      self.managed_services_docker_compose = managed_services_docker_compose
      self.external_service_docker_compose = external_service_docker_compose
      self.local_service_docker_compose = local_service_docker_compose
      self.assets_service_docker_compose = assets_service_docker_compose

    @pytest.fixture(scope="class", autouse=True)
    def create_scaffold_setup(self, runner, app_name, app_dir_path, target_workflow_name, external_service_docker_compose, local_service_docker_compose, assets_service_docker_compose, mock_data_directory_path, assets_service_mock_docker_compose_path):

      ScaffoldCliInvoker.cli_app_create(runner, app_dir_path, app_name)

      # Create external user defined service
      ScaffoldCliInvoker.cli_service_create(runner, app_dir_path, external_service_docker_compose.hostname, external_service_docker_compose.service_name, False)
      # Create local user defined services
      ScaffoldCliInvoker.cli_service_create(runner, app_dir_path, local_service_docker_compose.hostname, local_service_docker_compose.service_name, False)
      ScaffoldCliInvoker.cli_service_create_assets(runner, app_dir_path, assets_service_docker_compose.hostname, assets_service_docker_compose.service_name, False)

      # Don't run the local user defined service in the 'test' workflow
      # So don't copy the Docker Compose file over

      # Add user defined Docker Compose file for the assets service
      destination_path = Path(assets_service_docker_compose.docker_compose_path)
      assert destination_path.is_file()
      shutil.copyfile(assets_service_mock_docker_compose_path, destination_path)

      ScaffoldCliInvoker.cli_service_create_assets(runner, app_dir_path, assets_service_docker_compose.hostname, assets_service_docker_compose.service_name, False)

      # Add assets for assets service
      data_dir_path = DataDir.instance().path
      destination_assets_path = f"{data_dir_path}/docker/{assets_service_docker_compose.service_name}/{target_workflow_name}/index.html"
      destination_path = Path(destination_assets_path)
      assets_mock_path = mock_data_directory_path / "scaffold" / "index.html"
      shutil.copyfile(assets_mock_path, destination_path)

      # Created shared file in fixtures folder
      app = App(app_dir_path, DOCKER_NAMESPACE)
      config = {
        'workflow_name': target_workflow_name,
        'service_name': external_service_docker_compose.service_name 
      }
      command = ServiceWorkflowValidateCommand(app, **config)
      with open(f"{command.public_dir_path}/shared_file.txt", 'w') as file:
        file.write('this is a shared file')

      ScaffoldCliInvoker.cli_workflow_up(runner, app_dir_path, target_workflow_name=target_workflow_name)
    
    @pytest.fixture(scope="class", autouse=True)
    def cleanup_after_all(self, runner, app_dir_path, target_workflow_name):
      yield
      ScaffoldCliInvoker.cli_workflow_down(runner, app_dir_path, target_workflow_name)
      shutil.rmtree(app_dir_path)

    def test_no_core_services(self, app_dir_path, target_workflow_name):
      app = App(app_dir_path, DOCKER_NAMESPACE)
      config = {
        'workflow_name': target_workflow_name,
        'service_name': 'build'
      }

      command = WorkflowValidateCommand(app, **config)
      command.validate()

    def test_user_services(self, app_dir_path, target_workflow_name, external_service_docker_compose, local_service_docker_compose):
      app = App(app_dir_path, DOCKER_NAMESPACE)

      config = {
        'workflow_name': target_workflow_name,
        'service_name': external_service_docker_compose.service_name 
      }
      command = ServiceWorkflowValidateCommand(app, **config)
      command.validate()

      config = {
        'workflow_name': target_workflow_name,
        'service_name': local_service_docker_compose.service_name 
      }
      command = ServiceWorkflowValidateCommand(app, **config)
      command.validate()

    def test_tests(self):
      # This is covered by test_assets
      pass

    def test_assets(self, app_dir_path, target_workflow_name):
      app = App(app_dir_path, DOCKER_NAMESPACE)
      config = {
        'workflow_name': target_workflow_name,
        'service_name': 'assets'
      }

      command = ServiceWorkflowValidateCommand(app, **config)
      command.validate()


import pdb

from stoobly_agent.app.cli.helpers.iterate_group_by import iterate_group_by
from stoobly_agent.app.cli.helpers.replay_facade import ReplayCliOptions, ReplayFacade, TestCliOptions
from stoobly_agent.app.models.scenario_model import ScenarioModel
from stoobly_agent.app.proxy.replay.replay_scenario_service import inject_replay
from stoobly_agent.app.proxy.replay.trace_context import TraceContext
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import mode
from stoobly_agent.lib.api.keys import ProjectKey, ScenarioKey

class ScenarioFacade(ReplayFacade):

  def __init__(self, settings: Settings):
    super().__init__(settings)
    self.__model = ScenarioModel(settings)

  def create(self, project_key: str, name: str, description: str = ''):
    key = ProjectKey(project_key)
    return self.__model.create(**{
      'project_id': key.id, 
      'description': description,
      'name': name,
    })

  def index(self, project_key, cli_options: dict):
    key = ProjectKey(project_key)
    return self.__model.index(**{ 'project_id': key.id, **cli_options})

  def show(self, scenario_key: str):
    key = ScenarioKey(scenario_key)
    return self.__model.show(key.id)

  def replay(self, source_key: str, cli_options: ReplayCliOptions):
    return self.__replay(source_key, {
      'mode': mode.RECORD if cli_options.get('record') else mode.REPLAY,
      **self.common_replay_cli_options(cli_options),
      **self.__common_replay_options(source_key),
    })

  def snapshot(self, scenario_key: str, cli_options: dict):
    key = ScenarioKey(scenario_key)

    return self.__model.snapshot(key.id, **cli_options)

  def test(self, scenario_key: str, cli_options: TestCliOptions):
    return self.__replay(scenario_key, {
      'mode': mode.TEST,
      'report_key': cli_options.get('report_key'),
      'scenario_key': scenario_key, # Mock the request from the specified scenario instead of active scenario
      **self.common_test_cli_options(cli_options),
      **self.__common_replay_options(scenario_key),
    })

  def delete(self, scenario_key: str):
    key = ScenarioKey(scenario_key)

    return self.__model.destroy(key.id)

  def __common_replay_options(self, scenario_key: str):
    return {
      'project_key': ProjectKey.encode(ScenarioKey(scenario_key).project_id),
    }

  def __replay(self, scenario_key: str, replay_options: ReplayCliOptions):
    trace_context = replay_options['trace_context']
    group_by = replay_options['group_by']

    if not group_by or not trace_context:
      replay = inject_replay()
      return replay(scenario_key, replay_options)
    else:
      return iterate_group_by(
        group_by, 
        trace_context,
        lambda trace_context: self.__replay_with_trace_context(trace_context, scenario_key, replay_options)
      )

  def __replay_with_trace_context(self, trace_context: TraceContext, scenario_key: str, replay_options: ReplayCliOptions):
    _replay_options = replay_options.copy()
    _replay_options['trace_context'] = trace_context
    replay = inject_replay()
    return replay(scenario_key, _replay_options)



from typing import List, TypedDict, Union

from stoobly_agent.config.constants import test_strategy

from .pagination_query_params import PaginationQueryParams

class TestShowResponse(TypedDict):
  expected_latency: int
  id: int
  log: str
  passed: bool
  project_id: str
  skipped: bool
  strategy: Union[test_strategy.DIFF, test_strategy.CUSTOM, test_strategy.FUZZY] 

class TestsIndexQueryParams(PaginationQueryParams):
  project_id: str
  report_id: str
  scenario_id: str

class TestCreateParams(TypedDict):
  log: str
  passed: bool
  report_id: str
  request: bytes
  request_id: str
  status: str
  strategy: str

class TestsIndexResponse(TypedDict):
  list: List[TestShowResponse]
  total: int


import pdb

from typing import Iterable, List, TypedDict, Union

from stoobly_agent.app.proxy.test.context_abc import TestContextABC as TestContext
from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.config.constants.lifecycle_hooks import ON_LENGTH_MATCH_ERROR, ON_PARAM_NAME_EXISTS_ERROR, ON_PARAM_NAME_MISSING_ERROR, ON_TYPE_MATCH_ERROR, ON_VALID_TYPE_ERROR, ON_VALUE_MATCH_ERROR
from stoobly_agent.lib.api.interfaces.endpoints import RequestComponentName, ResponseParamName
from stoobly_agent.lib.utils.python_to_ruby_type import convert

def build_match_context(context: TestContext, facade: RequestComponentNamesFacade):
    return MatchContext({ 
        'lifecycle_hooks': context.lifecycle_hooks,
        'path_key': '', 
        'query': '', 
        'request_component_names_facade': facade
    })

class IMatchContext(TypedDict):
    lifecycle_hooks: dict
    path_key: str
    query: str
    request_component_names_facade: RequestComponentNamesFacade

class MatchContext():

    def __init__(self, context: IMatchContext):
        self.keys = context.get('keys') or [] # List of visited keys
        self.path_key = context['path_key']
        self.query = context['query']

        self.__lifecycle_hooks = context.get('lifecycle_hooks') # Optional
        self.__request_component_names_facade = context['request_component_names_facade']

    @property
    def current_key(self):
        length = len(self.keys)

        if length == 0:
            return
        return self.keys[length - 1]

    @property
    def request_component_names_query_index(self):
        return self.__request_component_names_facade.query_index

    @property
    def request_component_name(self) -> Union[RequestComponentName, None]:
        return self.request_component_names_query_index.get(self.query)

    @property
    def children(self) -> List[RequestComponentName]:
        if self.query == '':
            return self.__request_component_names_facade.edges_index.get(None) or []

        component_names = self.request_component_names_query_index.get(self.query)
        if not component_names:
            return []

        if not isinstance(component_names, list):
            component_names = [component_names]

        edges_index = self.__request_component_names_facade.edges_index

        for component_name in component_names:
            edges = edges_index.get(component_name.get('id'))
            if edges:
                return edges

        return []

    def to_dict(self) -> IMatchContext:
        return {
            'lifecycle_hooks': self.__lifecycle_hooks,
            'keys': self.keys.copy(),
            'path_key': self.path_key,
            'query': self.query,
            'request_component_names_facade': self.__request_component_names_facade,
        }

    def clone(self):
        return __class__(self.to_dict())

    def visit_list(self, key):
        self.keys.append(key)
        self.path_key = f"{self.path_key}[{key}]"
        self.query = f"{self.query}[*]"

    def visit_dict(self, key):
        self.keys.append(key)
        self.path_key = '.'.join([self.path_key, key]) if len(self.path_key) > 0 else key
        self.query = '.'.join([self.query, key]) if len(self.query) > 0 else key

    def selected(self):
        return self.__request_component_names_facade.is_selected(self.query)

    def ignored(self, expected_value = None, actual_value = None):
        return not self.selected() or (not self.required() and self.__required_matches(expected_value, actual_value))

    def deterministic(self) -> bool:
        request_component_names_facade: RequestComponentNamesFacade = self.__request_component_names_facade
        if not request_component_names_facade or len(request_component_names_facade.all) == 0:
            return True

        query: str = self.query
        deterministic_param_names: ResponseParamName = request_component_names_facade.deterministic
        return self.__param_name_matches(query, deterministic_param_names)

    def required(self) -> bool:
        request_component_names_facade: RequestComponentNamesFacade = self.__request_component_names_facade
        if not request_component_names_facade or len(request_component_names_facade.all) == 0:
            return True

        query: str = self.query
        required_param_names: ResponseParamName = request_component_names_facade.required
        return self.__param_name_matches(query, required_param_names)

    ### Matchers

    def length_matches(self, v1, v2):
        return len(v1) == len(v2)

    def param_name_exists(self, key, actual):
        return key in actual

    def value_contract_matches(self, v1, contracts: Union[RequestComponentName, List[RequestComponentName]]):
        if not contracts:
            return True

        if not isinstance(contracts, list):
            contracts = [contracts]

        v_type = str(type(v1))
        for contract in contracts:
            if not contract.get('inferred_type'):
                return True

            if convert(v_type) == contract.get('inferred_type'):
                return True

        return False

    def value_fuzzy_matches(self, v1, v2):
        return type(v1) == type(v2)

    def value_matches(self, v1, v2):
        return v1 == v2

    def value_type_exists(self, value, valid_types: list):
        return type(value) in valid_types

    def value_type_matches(self, t1, t2):
        if type(t1) == type(t2):
            return True

        if t1 == dict or t2 == dict:
            return self.value_is_dict(t1) and self.value_is_dict(t2)

        if t1 == list or t2 == list:
            return self.value_is_list(t1) and self.value_is_list(t2)

        return False

    def value_is_dict(self, d):
        if isinstance(d, dict):
            return True
        return hasattr(d, 'items') and isinstance(d, Iterable)

    def value_is_list(self, l):
        if isinstance(l, list):
            return True
        return hasattr(l, 'sort') and isinstance(l, Iterable)

    ### Errors

    def handle_length_match_error(self, actual):
        handler = self.__lifecycle_hook(ON_LENGTH_MATCH_ERROR)

        if handler:
            return handler(self.clone(), actual)

    def handle_param_name_exists_error(self, actual):
        handler = self.__lifecycle_hook(ON_PARAM_NAME_EXISTS_ERROR)

        if handler:
            return handler(self.clone(), actual)

    def handle_param_name_missing_error(self, actual):
        handler = self.__lifecycle_hook(ON_PARAM_NAME_MISSING_ERROR)

        if handler:
            return handler(self.clone(), actual)

    def handle_type_match_error(self, actual):
        handler = self.__lifecycle_hook(ON_TYPE_MATCH_ERROR)

        if handler:
            return handler(self.clone(), actual)

    def handle_valid_type_error(self, actual):
        handler = self.__lifecycle_hook(ON_VALID_TYPE_ERROR)

        if handler:
            return handler(self.clone(), actual)
    
    ### Private

    def __lifecycle_hook(self, name):
        if not self.__lifecycle_hooks:
            return

        return self.__lifecycle_hooks.get(name)

    def __param_name_matches(self, query, param_names: List[ResponseParamName]) -> bool:
        for param_name in param_names:
            # Not all components will have a query property,
            # If query property does not exist, then use name property
            _query = param_name['query'] if 'query' in param_name else param_name['name']

            if _query == query:
                return True

        return False

    def __required_matches(self, v1, v2):
        if v1 == None or v2 == None:
            return v1 == v2
        return True


from typing import List

from .types.proxy_settings import UrlRule as IUrlRule

class UrlRule:

  def __init__(self, url_rule: IUrlRule):
    self.update(url_rule)

  @property
  def hostname(self):
    return self.__hostname

  @property 
  def modes(self):
    return self.__modes

  @property
  def path(self):
    return self.__path

  @property
  def port(self):
    return self.__port

  @property
  def scheme(self):
    return self.__scheme

  def update(self, url_rule: IUrlRule):
    self.__url_rule = url_rule
    self.__hostname = self.__url_rule.get('hostname')
    self.__modes = self.__url_rule.get('modes')
    self.__path = self.__url_rule.get('path')
    self.__port = self.__url_rule.get('port')
    self.__scheme = self.__url_rule.get('scheme')

  def to_dict(self):
    return {
      'hostname': self.__hostname,
      'modes': self.__modes,
      'path': self.__path,
      'port': self.__port,
      'scheme': self.__scheme,
    }


import json
import pdb
import pytest
import requests

from click.testing import CliRunner

from stoobly_agent.test.test_helper import reset

from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.response import Response

from stoobly_agent.app.models.factories.resource.local_db.helpers.request_builder import RequestBuilder
from stoobly_agent.app.models.factories.resource.local_db.request_adapter import LocalDBRequestAdapter
from stoobly_agent.app.models.factories.resource.local_db.request_adapter import LocalDBResponseAdapter

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestOrmPythonResponseAdapter():
  @pytest.fixture(scope='class')
  def settings(self):
    return reset()

  @pytest.fixture(scope='class')
  def response_body(self):
    return {
      'first_name': 'John',
      'last_name': 'Smith',
    }

  @pytest.fixture(scope='class')
  def response_headers(self):
    return {
      'Content-Type': 'application/json'
    }

  @pytest.fixture(scope='class')
  def local_db_request_adapter(self):
    adapter = LocalDBRequestAdapter(request_orm=Request, response_orm=Response)
    return adapter

  @pytest.fixture(scope='class')
  def local_db_response_adapter(self):
    adapter = LocalDBResponseAdapter(request_orm=Request)
    return adapter

  @pytest.fixture(scope='function')
  def created_response(
    self, settings: Settings, 
    response_headers: dict, response_body: dict
  ):
    status = RequestBuilder(
      method='POST',
      request_body='',
      response_body=response_body,
      response_headers=response_headers,
      status_code=200,
      url='https://petstore.swagger.io/v2/pets',
    ).with_settings(settings).build()[1]
    assert status == 200

    return Response.last()

  def test_it_updates_response(self, created_response: Response, local_db_response_adapter: LocalDBResponseAdapter):
    new_body = '123'
    local_db_response_adapter.update(created_response.id, text=new_body)

    updated_response, status = local_db_response_adapter.show(created_response.id)

    assert status == 200
    assert updated_response[0]['text'] == new_body
    


from .joined_request_adapter import JoinedRequestAdapter
from .raw_http_request_adapter import RawHttpRequestAdapter
from .raw_http_response_adapter import RawHttpResponseAdapter
from .raw_joined import RawJoinedRequestAdapterFactory


import pdb

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource

def inject_search_endpoint(intercept_settings: InterceptSettings):
  remote = intercept_settings.settings.remote
  resource = EndpointsResource(remote.api_url, remote.api_key)
  return lambda project_id, method, url: search_endpoint(resource, project_id, method, url)

def search_endpoint(endpoints_resource: EndpointsResource, project_id: int, method: str, url: str, **query_params):
  res = endpoints_resource.index(
    **{
      'method': method,
      'project_id': project_id,
      'q': url,
      'size': 1,
    },
    **query_params
  )

  if res.status_code != 200:
    return None

  endpoints = res.json()

  if len(endpoints) == 0:
    return None

  return endpoints[0]



import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import reset

from stoobly_agent.cli import config, intercept, scenario
from stoobly_agent.lib.api.keys import ScenarioKey
from stoobly_agent.lib.orm.scenario import Scenario


from stoobly_agent.config.constants import mode, record_policy

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

@pytest.fixture(autouse=True, scope='module')
def settings():
  return reset()

class TestInterceptConfigure():

  class TestRecordPolicy():

    class TestScenario():
      @pytest.fixture(scope='class')
      def scenario_name(self):
        return 'Login 1.0.0'

      @pytest.fixture(scope='class', autouse=True)
      def scenario(self, runner: CliRunner, scenario_name: str):
        scenario_create_result = runner.invoke(scenario, ['create', '--without-headers', '--select', 'key', scenario_name])
        assert scenario_create_result.exit_code == 0

        return Scenario.last()

      @pytest.fixture(scope='class', autouse=True)
      def configure(self, runner: CliRunner, scenario: Scenario):
        configure_result = runner.invoke(config, ['scenario', 'set', scenario.key()])
        assert configure_result.exit_code == 0

        configure_result = runner.invoke(intercept, ['configure', '--mode', mode.RECORD ,'--policy', record_policy.OVERWRITE])
        assert configure_result.exit_code == 0

      class TestChangingPolicy():

        def test_scenario_not_overwritable(self, runner: CliRunner, scenario: Scenario):
          Scenario.find(scenario.id).update(overwritable=True)

          configure_result = runner.invoke(intercept, ['configure', '--policy', record_policy.ALL])
          assert configure_result.exit_code == 0

          assert not Scenario.find(scenario.id).overwritable



import os

from stoobly_agent.config.constants import env_vars

from .types.remote_settings import RemoteSettings as IRemoteSettings

class RemoteSettings:

  def __init__(self, remote_settings: IRemoteSettings):
    self.__remote_settings = remote_settings

    self.__api_key = self.api_key_before_change
    self.__api_url = self.api_url_before_change
    self.__project_key = self.project_key_before_change

  @property
  def api_key_before_change(self):
    return self.__remote_settings.get('api_key')

  @property
  def api_key(self) -> str:
    def get():
      if self.__api_key != self.api_key_before_change:
        return self.__api_key

      if os.environ.get(env_vars.API_KEY):
        return os.environ[env_vars.API_KEY]

      return self.__api_key
    
    _api_key = get()
    return _api_key.strip() if isinstance(_api_key, str) else ''

  @api_key.setter
  def api_key(self, v: str):
    self.__api_key = v.strip()

  @property
  def api_url_before_change(self):
    return self.__remote_settings.get('api_url')

  @property
  def api_url(self):
    if self.__api_url != self.api_url_before_change:
      return self.__api_url

    if os.environ.get(env_vars.API_URL):
      return os.environ[env_vars.API_URL]

    return self.__api_url

  @property
  def project_key_before_change(self):
    return self.__remote_settings.get('project_key')

  @property
  def project_key(self):
    if self.__project_key != self.project_key_before_change:
      return self.__project_key

    if os.environ.get(env_vars.AGENT_REMOTE_PROJECT_KEY):
      return os.environ[env_vars.AGENT_REMOTE_PROJECT_KEY]
      
    return self.__project_key

  @project_key.setter
  def project_key(self, v):
    self.__project_key = v

  @api_url.setter
  def api_url(self, v):
    self.__api_url = v

  def to_dict(self) -> IRemoteSettings:
    return {
      'api_key': self.__api_key,
      'api_url': self.__api_url,
      'project_key': self.__project_key,
    }


from mitmproxy.http import Request as MitmproxyRequest
from typing import Union
from ...config.constants import mock_policy, record_policy

from stoobly_agent.config.constants import custom_headers
from stoobly_agent.app.settings import Settings
from stoobly_agent.app.settings.types import IProjectModeSettings

from ...config.constants import mode

def get_project_key(headers: MitmproxyRequest.headers, settings: IProjectModeSettings) -> str:
    if custom_headers.PROJECT_KEY in headers:
        return headers[custom_headers.PROJECT_KEY]

    return settings.get('project_key')

def get_report_key(headers: MitmproxyRequest.headers) -> Union[str, None]:
    return headers.get(custom_headers.REPORT_KEY)

def get_scenario_key(headers: MitmproxyRequest.headers, settings: IProjectModeSettings) -> str:
    if custom_headers.SCENARIO_KEY in headers:
        return headers[custom_headers.SCENARIO_KEY]

    return settings.get('scenario_key')

def is_proxy_enabled(headers: MitmproxyRequest.headers, settings: IProjectModeSettings) -> bool:
    return settings.get('enabled') or custom_headers.PROXY_MODE in headers

def get_proxy_mode(headers: MitmproxyRequest.headers, settings: Settings) -> str:
    access_control_header =  'Access-Control-Request-Headers'
    do_proxy_header = custom_headers.DO_PROXY

    if access_control_header in headers and do_proxy_header.lower() in headers[access_control_header]:
        return mode.NONE
    elif do_proxy_header in headers:
        return mode.NONE
    elif custom_headers.PROXY_MODE in headers:
        return headers[custom_headers.PROXY_MODE]
    else:
        return settings.proxy.intercept.active

def get_mock_policy(headers: MitmproxyRequest.headers, settings: IProjectModeSettings) -> str:
    if custom_headers.MOCK_POLICY in headers:
        return headers[custom_headers.MOCK_POLICY]
    else:
        return settings.get('policy') or mock_policy.FOUND

def get_record_policy(headers: MitmproxyRequest.headers, settings: Settings) -> str:
    if custom_headers.RECORD_POLICY in headers:
        return headers[custom_headers.RECORD_POLICY]
    else:
        return settings.get('policy') or record_policy.ALL


import pdb
import uuid
import time

from typing import Callable, TypedDict

from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.scenario import Scenario

from .request_snapshot import RequestSnapshot
from .scenario_snapshot import ScenarioSnapshot
from .snapshot_types import Action, DELETE_ACTION, PUT_ACTION, REQUEST_RESOURCE, Resource, SCENARIO_RESOURCE

COLUMN_DELIMITTER = ' '

class EventHandlers(TypedDict):
  handle_request_delete: Callable[[str], None]
  handle_request_put: Callable[[str], None]
  handle_scenario_delete: Callable[[str], None]
  handle_scenario_put: Callable[[str], None]

class LogEvent():
  def __init__(self, event: str):
    toks = event.strip().split(COLUMN_DELIMITTER)

    self.uuid = toks[0]
    self.resource = toks[1]
    self.resource_uuid = toks[2]
    self.action = toks[3]
    self.created_at = int(toks[4]) if len(toks) > 4 else None # milliseconds

  @classmethod
  def from_resource(cls, resource: Resource, action: Action):
    return cls(cls.serialize_resource(resource, action))

  @staticmethod
  def serialize(resource_name: Resource, resource_uuid: str, action: Action):
    return f"{str(uuid.uuid1())} {resource_name} {resource_uuid} {action} {int(time.time() * 1000)}"

  @classmethod
  def serialize_delete(cls, resource: Resource):
    return cls.serialize_resource(resource, DELETE_ACTION)

  @classmethod
  def serialize_put(cls, resource: Resource):
    return cls.serialize_resource(resource, PUT_ACTION)

  @classmethod
  def serialize_resource(cls, resource: Resource, action: Action):
    resource_name = ''

    if isinstance(resource, Request):
      resource_name = REQUEST_RESOURCE
    elif isinstance(resource, Scenario):
      resource_name = SCENARIO_RESOURCE
    else:
      resource_name = resource.__class__.__name__

    return cls.serialize(resource_name, resource.uuid, action)

  @property
  def key(self):
    return f"{self.resource_uuid}.{self.resource}"

  def __str__(self):
    return COLUMN_DELIMITTER.join([self.uuid, self.resource, self.resource_uuid, self.action, str(self.created_at)])

  def apply(self, **kwargs: EventHandlers):
    if self.is_request():
      if self.action == DELETE_ACTION:
        return kwargs['handle_request_delete'](self.resource_uuid)
      elif self.action == PUT_ACTION:
        return kwargs['handle_request_put'](self.resource_uuid)
    elif self.is_scenario():
      if self.action == DELETE_ACTION:
        return kwargs['handle_scenario_delete'](self.resource_uuid)
      elif self.action == PUT_ACTION:
        return kwargs['handle_scenario_put'](self.resource_uuid)

  def duplicate(self):
    event_str = str(self) 
    toks = event_str.split(COLUMN_DELIMITTER)
    toks[0] = str(uuid.uuid1())
    toks[4] = str(int(time.time() * 1000))
    return LogEvent(COLUMN_DELIMITTER.join(toks))

  def is_request(self):
    return self.resource == REQUEST_RESOURCE

  def is_scenario(self):
    return self.resource == SCENARIO_RESOURCE

  def snapshot(self):
    if self.is_request():
      return RequestSnapshot(self.resource_uuid)
    elif self.is_scenario():
      return ScenarioSnapshot(self.resource_uuid)

  def to_dict(self):
    return {
      'action': self.action,
      'created_at': self.created_at,
      'resource': self.resource,
      'resource_uuid': self.resource_uuid,
      'uuid': self.uuid,
    }


import os
import pdb

from stoobly_agent.config.constants import env_vars
from stoobly_agent.config.constants import mode

from .types.proxy_settings import InterceptSettings as IInterceptSettings

class InterceptSettings:
  __intercept_settings = None

  def __init__(self, intercept_settings: IInterceptSettings):
    self.__intercept_settings = intercept_settings or {}

    self.__active = self.__intercept_settings.get('active')
    self.__mode = self.mode_before_change
    self.__project_key = self.__intercept_settings.get('project_key')
    self.__upstream_url = self.__intercept_settings.get('upstream_url')

  @property
  def active(self):
    return self.__active

  @active.setter
  def active(self, v):
    self.__active = bool(v)
    self.__intercept_settings['active'] = bool(v)

  @property
  def mode_before_change(self):
    return self.__intercept_settings.get('mode') or mode.NONE

  @property
  def mode(self):
    if self.__mode != self.mode_before_change:
      return self.__mode

    if os.environ.get(env_vars.AGENT_ACTIVE_MODE):
        return os.environ[env_vars.AGENT_ACTIVE_MODE]
        
    return self.__mode

  @mode.setter
  def mode(self, v):
    if v in [mode.MOCK, mode.NONE, mode.RECORD, mode.REPLAY, mode.TEST]:
      self.__mode = v
      self.__intercept_settings['mode'] = v

  @property
  def project_key(self):
    return self.__project_key

  @project_key.setter
  def project_key(self, v):
    self.__project_key = v
    self.__intercept_settings['project_key'] = v

  @property
  def upstream_url(self):
    return self.__upstream_url

  def to_dict(self) -> IInterceptSettings:
    _dict = {
      'active': self.__active,
      'mode': self.__mode,
      'project_key': self.__project_key,
    }

    if self.__upstream_url:
      _dict['upstream_url'] = self.__upstream_url
    
    return _dict
